\section { A module for fuzzy controllers }
\label{SecModFContr}
With the relational model and the operations on fuzzy controllers developed in the last %%@
section, we are now able to provide a module for handling controllers. With this we aim %%@
at suitable combinators to construct them and test their behavior in an algebraic (i.e., %%@
essentially component-free) manner. We then have the necessary background to develop a %%@
graphical user interface (GUI) to make the module
\begin{code}
module FContr where

import LFuzzyRel
import LFuzzyRelCategories
import LFuzzyRelLattices
import Goguen
import Lattice
import List (elemIndex,nub,intersect,(\\))
\end{code}
comfortable to use. 

\subsubsection* { The data structures }
Essentially, we need two data structures --- one for linguistic entities and one for %%@
fuzzy controllers. Since linguistic entities are single morphisms of the underlying %%@
Goguen category and it 
plays no role how they were created (by ordering-based weakening/strengthening, shifting %%@
etc.), the 
most simple variant to represent them is chosen.
\begin{code}
data LingEntity mor  = LingEntity { entLabel :: String
                                   ,entRel   :: mor } 
\end{code}
The only thing we have to add is a label for each entity. These labels are important to %%@
model the 
rulebase later on. Throughout this module we will use instances of type %%@
\smalltt{LingEntity (FRel e 
obj obj)} of the data structure above.\\
Furthermore, the function 
\begin{code}
entTrg :: LingEntity (FRel e obj1 obj2) -> [obj2]
entTrg = trg . entRel 
\end{code}
which obviously gives back the target object of the linguistic entity, is useful.\\
The \smalltt{Eq} and \smalltt{Show} instances are implemented as follows.
\begin{code}
instance Eq (LingEntity mor) where
   x1 == x2 = entLabel x1 == entLabel x2
   
instance (Show mor) => Show (LingEntity mor) where
   show le = "ENTITY NAME     : "++entLabel le++"\n"++
             "ENTITY RELATION : "++"\n" ++ show (entRel le)
\end{code}
Notice that the user has to take care to avoid duplicate entity names since they are the %%@
only 
equality criteria of the \smalltt{Eq} instance. We prefer this variant due to efficiency 
considerations. 

From Definition \ref{DefnSimpContr} we know that a simple controller includes a %%@
fuzzification, application, aggregation and defuzzification function. Some of these %%@
parameters are restricted to be equal with two fuzzy controllers that shall be put %%@
together by one of the introduced operations. This implies that we will have to check %%@
them for equality at some point. Hence, they have to be labeled analogously to the %%@
linguistic entities.
\begin{code}
data UnFunc mor  = UnFunc { unLabel :: String,
                            unFunc  :: mor -> mor }
data BinFunc mor = BinFunc { binLabel :: String,
                             binFunc  :: mor -> mor -> mor }
data DerOp e     = DerOp { opLabel :: String,
                           opLoos  :: Loos e,
                           comm    :: Bool,
                           idem    :: Bool }
\end{code}
Obviously, we have structures for unary functions \smalltt{UnFunc} (fuzzification, %%@
defuzzification), binary functions \smalltt{BinFunc} (application) and derived operations %%@
\smalltt{DerOp} (aggregation). The structure for derived operations carries two %%@
additional flags to determine whether the loos is commutative (\smalltt{comm}) and %%@
idempotent (\smalltt{idem}), respectively. This helps to compute the rulebase more %%@
efficiently, later on. Notice that we also restricted the application function to be a %%@
composition-based derived operation. But, we have to generate a new application function, %%@
if we apply one of the operations $\wh{+_{\sqcup_{\oplus}}}$ resp.\ %%@
$\wh{\times_{\sqcap_{\odot}}}$ to given controllers. Hence, \smalltt{DerOp} would not %%@
suffice to model this.\\
The induced \smalltt{Show} and \smalltt{Eq} instances are implemented as follows.
\begin{code}
instance Eq (DerOp e) where
   op1 == op2 = opLabel op1 == opLabel op2
instance Eq (UnFunc mor) where
   f1 == f2 = unLabel f1 == unLabel f2
instance Eq (BinFunc mor) where
   f1 == f2 = binLabel f1 == binLabel f2

instance Show (DerOp e) where
   show = show . opLabel
instance Show (UnFunc mor) where
   show = show . unLabel
instance Show (BinFunc mor) where
   show = show . binLabel
\end{code}
Now, we are ready to focus on the controller data structure. Our target is a GUI for a %%@
comfortable creation of fuzzy controllers. Therefore, it is essential that the data %%@
structures make some kind of rollback possible. It has to carry the history of creation %%@
so that the user can step back to a previous state without losing information. This %%@
induces a recursive data structure. From Figures \ref{Contr2} and \ref{Contr3} we know %%@
that a fuzzy controller may form the sum or the product of several subcontrollers. Hence, %%@
we need a variant record covering three cases --- a simple controller, a sum controller %%@
and a product controller.
\begin{code}
type Rule           e   = (DerOp e,(String,[String]))

data Header   e obj mor = Header { contrLabel :: String      -- controller label
                                  ,unitObj    :: [obj]       -- unit object
                                  ,entryLat   :: Lat e       -- entry lattice
                                  ,fuzz       :: UnFunc mor  -- fuzzification
                                  ,appl       :: BinFunc mor -- application
                                  ,defuzz     :: UnFunc mor} -- defuzzification
                                  
data FuzContr e obj = SimpleContr { header     :: Header e obj (FRel e obj obj)
                                   ,aggr       :: DerOp e       -- aggregation
                                   ,rules      :: [Rule e]      -- rulebase
                                   ,lingEntIn  :: [LingEntity (FRel e obj obj)]
                                   ,lingEntOut :: [LingEntity (FRel e obj obj)]}
                      | SumContr  { header     :: Header e obj (FRel e obj obj)
                                   ,srcSum     :: Bool    
                                   ,trgSum     :: Bool
                                   ,derJoin    :: DerOp e
                                   ,subContrs  :: [FuzContr e obj]}
                      | ProdContr { header     :: Header e obj (FRel e obj obj)
                                   ,srcProd    :: Bool    
                                   ,trgProd    :: Bool
                                   ,derMeet    :: DerOp e
                                   ,subContrs  :: [FuzContr e obj]}
\end{code}
The data structures \smalltt{Rule} (for a single rule ) and \smalltt{Header} are pretty %%@
intuitive. A rule consists of an interpretation function (\smalltt{DerOp}), an input %%@
entity and a list of output entities it shall be connected with. A list of such rules %%@
then forms the rulebase of a controller. For example, the
rulebase
\begin{tabbing}
\qquad \= if $x$ is $Q_{1}$ then $y$ is $S_{1}$ \\
       \> if $x$ is $Q_{2}$ then $y$ is $S_{3}$ and $S_{4}$
\end{tabbing}
has to be represented by the list \\
\centerline{\smalltt{[({\em <op1>},("Q1",["S1"])),({\em <op2>},("Q2",["S3","S4"]))]}}. %%@
Notice that the arrangement of the list entries plays no role if the aggregation function %%@
is commutative.\\
The header is, in wide areas, a realization of the needed functionality to infer the %%@
output. Thus, the header is needed by all three kinds of controller. As an extension to %%@
Definition \ref{DefnSimpContr}, the header includes a label for the controller %%@
(\smalltt{contrLabel}), the underlying entry lattice (\smalltt{entryLat}) of the %%@
$\SL$-fuzzy relations and a unit object (\smalltt{unitObj}). The last two parameters are %%@
needed to automatically construct the underlying Goguen category, later on.\\
With these preparations, the \smalltt{FuzContr} data structure is straightforward. The %%@
simple controller \smalltt{SimpleContr}, in addition to the header, gets all needed %%@
functions to create the core. This includes an aggregation function \smalltt{aggr}, a %%@
rulebase \smalltt{rules} and the linguistic entities for the input (\smalltt{lingEntIn}) %%@
and output (\smalltt{lingEntOut}), respectively. Notice that \smalltt{Header} is %%@
instantiated using \smalltt{FRel e obj obj} for the type parameter. This implies that the %%@
modules \smalltt{FContr} and \smalltt{LFuzzyRel} can only be used together. A more %%@
general approach seems not to be senseful since too much specific functionality of %%@
\smalltt{LFuzzyRel} is used.\\
The data structures \smalltt{SumContr} and \smalltt{ProdContr} are implemented equally. %%@
Besides the header, they get two flags which indicate, whether the direct sum resp.\ %%@
cross product shall be generated with the source resp.\ target of the controller. The %%@
meaning of these two flags is summarized in Table \ref{TabOpFuzContr}.
\begin{table}[h]
\label{TabOpFuzContr}
\centering
\begin{tabular}{|ll|l|l|}
%Row: 1
\cline{1-4}
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to7,80ex{\hfil \smalltt{sS}/\smalltt{sP}\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to7,40ex{\hfil \smalltt{tS}/\smalltt{tP}\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to20,40ex{\hfil Operation \smalltt{SumContr}\hfil}} %%@
& 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to20,20ex{\hfil Operation %%@
\smalltt{ProdContr}\hfil}} \\
%Row: 2
\cline{1-4}
\vbox to3,88ex{\vspace{1pt}\vfil\hbox to7,80ex{\hfil \smalltt{True}\hfil}} & 
\vbox to3,88ex{\vspace{1pt}\vfil\hbox to7,40ex{\hfil \smalltt{True}\hfil}} & 
\vbox to3,88ex{\vspace{1pt}\vfil\hbox to20,40ex{\hfil $\wh{+_{\sqcup_{\oplus}}}$\hfil}} & 
\vbox to3,88ex{\vspace{1pt}\vfil\hbox to20,20ex{\hfil %%@
$\wh{\times_{\sqcup_{\odot}}}$\hfil}} \\
%Row: 3
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to7,80ex{\hfil \smalltt{True}\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to7,40ex{\hfil \smalltt{False}\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to20,40ex{\hfil %%@
$\wh{+_{\sqcup_{\oplus}}}^{s}$\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to20,20ex{\hfil %%@
$\wh{\times_{\sqcup_{\odot}}}^{s}$\hfil}} \\
%Row: 4
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to7,80ex{\hfil \smalltt{False}\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to7,40ex{\hfil \smalltt{True}\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to20,40ex{\hfil %%@
$\wh{+_{\sqcup_{\oplus}}}^{t}$\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to20,20ex{\hfil %%@
$\wh{\times_{\sqcup_{\odot}}}^{t}$\hfil}} \\
%Row: 5
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to7,80ex{\hfil \smalltt{False}\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to7,40ex{\hfil \smalltt{False}\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to20,40ex{\hfil $\sqcup_{\oplus}$\hfil}} & 
\vbox to2,88ex{\vspace{1pt}\vfil\hbox to20,20ex{\hfil $\sqcup_{\odot}$\hfil}} \\
\cline{1-4}
\end{tabular}
\caption { The meaning of the flags \smalltt{sS}/\smalltt{tS} and %%@
\smalltt{sP}/\smalltt{tP} }
\end{table}\\
Furthermore, both \smalltt{SumContr} and \smalltt{ProdContr} carry a $\sqcap$-based %%@
derived operation \smalltt{derJoin} resp.\ \smalltt{derMeet}. Notice, that the %%@
controllers in the subcontroller list \smalltt{subctrs} have to be combinable (cf.\ %%@
Definition \ref{DefnCombinable}) and \smalltt{derJoin} / \smalltt{derMeet} have to %%@
fulfill Definition  \ref{DefnCombine} to make  all operations of the table above %%@
applicable.\\
Finally, we want to mention that the information the header carries could be extracted %%@
form the subcontroller list with \smalltt{SumContr} and \smalltt{ProdContr}. Hence, the %%@
header would not be necessary for these structures. But, we prefer this variant to have %%@
quick access to the parameters. Since no really high depth of the \smalltt{FuzContr} data %%@
structure is to expect, the resulting overhead should be acceptable. Real applications %%@
will have to show whether this approach suffices or changes have to be made. Furthermore, %%@
the variants \smalltt{SumContr} and \smalltt{ProdContr} could have been put together to a %%@
single variant \smalltt{ExtendedContr} or similar using an additional flag to determine %%@
which operation is wanted. But again, we prefer to have separate structures for %%@
convenience and clearness.

Again, we provide \smalltt{Eq} and \smalltt{Show} instances.
\begin{code}
instance Eq (FuzContr e obj) where
   fc1 == fc2 = contrLabel (header fc1) == contrLabel (header fc2)

instance (Show obj, Show e) => Show (FuzContr e obj) where
   show (SimpleContr (Header l u e f ap d) ag rs lI lO) =
     let presentEnt = 
           concat . 
            map (\x -> "\n"++"    ENTITY NAME : "++ entLabel x++"\n"++
                             "    RELATION    : "++ "\n"++ show (entRel x))
     in
       "CONTROLLER NAME     : "++ l ++"\n"++
       "UNIT OBJECT         : "++ show u ++ "\n" ++
       "LING.ENTITUES INPUT : "++ presentEnt lI ++
       "LING.ENTITIES OUTPUT: "++ presentEnt lO ++ "\n"++
       "RULEBASE            : "++ (concat $ map (\x -> show x++"\n"++
                                                  "                      ") rs )
   show (SumContr  (Header l u _ _ _ _) _ _ _ subctrs) = 
     "CONTROLLER NAME     : "++ l ++"\n"++
     "UNIT OBJECT         : "++ show u ++"\n"++
     "SUBCONTROLLERS      : "++ concat (map (contrLabel.header) subctrs)
   show (ProdContr (Header l u e f ap d) dM sP tP subctrs) = 
           show (SumContr(Header l u e f ap d) dM sP tP subctrs)
\end{code}
As with \smalltt{LingEntity}, equality of two controllers is reduced to label equality. %%@
If the user explicitly wants to test whether the cores of two different controllers are %%@
equal, he has to use the function \smalltt{core} introduced later on.

\subsubsection* {Computing the core of a controller}
Having the necessary data structures, we now aim at suitable auxiliary functions to make %%@
the 
instantiation of fuzzy controllers and the computation of their cores comfortable for the %%@
user. With 
the next functions we want to support constructing the input resp.\ output linguistic %%@
entites. We introduce ordering and residual-based strengthening/weakening  as well as %%@
shifting to build up linguistic entities that fit together.
\begin{code}
greaterThan e m = fConv $ fLRes (fConv e) m
lessThan    e m = fConv $ fLRes e m 

very xi m 1    = fLRes  m xi
very xi m i    = fLRes (very xi m $ i-1) xi

roughly xi m 1 = fComp m xi
roughly xi m i = fComp (roughly xi m $ i-1) xi

shiftL m bij 1 = fComp m $ fConv bij
shiftL m bij i = fComp (shiftL m bij $ i-1) $ fConv bij

shiftR m bij 1 = fComp m bij
shiftR m bij i = fComp (shiftR m bij $ i-1) bij
\end{code}
\begin{sloppypar}
These routines are direct realizations of the underlying definitions. Notice that we %%@
delibe\-rately do not yet use the \smalltt{Goguen} module to compute the terms. %%@
Throughout this module we differentiate between two different levels of abstraction. %%@
Everything that has to be done to construct a controller is done on a level that is not %%@
pure algebraic. For these actions the modules \smalltt{LFuzzyRel} and \smalltt{Lattice} %%@
are used. But, all operations on controllers themselves are done using the underlying %%@
Goguen category, which is automatically generated. This has the further advantage that %%@
the user does not have to instantiate it himself.
\end{sloppypar}

The following functions help to compute the underlying Goguen category of a fuzzy %%@
controller.
\begin{code}
contrDom, contrRan :: FuzContr e obj -> [obj]
contrDom (SimpleContr _ _ _ lI _)  = if null lI then [] else entTrg $ head lI
contrDom (SumContr    _ sS _ _ subctrs) = 
    if sS then concat $ map contrDom subctrs
    else contrDom (head subctrs)
contrDom (ProdContr   _ sP _ _ subctrs) = 
    if sP then crossProd $ map contrDom subctrs 
    else contrDom (head subctrs)

contrRan (SimpleContr _ _ _ _ lO)  = if null lO then [] else entTrg $ head lO
contrRan (SumContr    _ _ tS _ subctrs) = 
    if tS then concat $ map contrRan subctrs
    else contrRan (head subctrs)
contrRan (ProdContr   _ _ tP _ subctrs) = 
    if tP then crossProd $ map contrRan subctrs 
    else contrRan (head subctrs)

catObjects :: (Eq obj) => FuzContr e obj -> [[obj]]
catObjects fc@(SimpleContr h _ _ _ _) =
    nub $ [unitObj h,contrDom fc,contrRan fc]
catObjects fc = 
    nub $ contrDom fc : contrRan fc : map (concat . catObjects) (subContrs fc)

contrCat :: (Eq e, Eq obj) => FuzContr e obj -> 
                              Gog (Loos e) [obj] (FRel e obj obj)
contrCat fc = let eLat = entryLat $ header fc in
                fRelGogCat eLat (getAllFRels eLat $ catObjects fc)
\end{code}
Obviously, \smalltt{contrDom} resp.\ \smalltt{contrRan} deliver the domain and range of %%@
the 
controller, respectively. To do so in the \smalltt{SimpleContr} case, they build up the %%@
direct sums 
by simple list concatenation. In the \smalltt{ProdContr} case, the function %%@
\smalltt{crossProd} is used to compute the relational products of the domains resp.\ %%@
ranges of all subcontrollers. The exact manner of function of this routine is explained %%@
below.\\
The necessary objects to instantiate the underlying Goguen category are delivered by 
\linebreak\smalltt{catObjects}. The simple case \smalltt{SimpleContr} provides the unit %%@
object, the domain and 
range of the controller and the targets of all input and output entities, respectively. %%@
With 
controllers of type \smalltt{ProdContr} we again have to add the products of the input %%@
resp.\ output 
spaces of the subcontrollers.

To understand the computation of the core of a product controller later on, it is very %%@
important to know how the function
\begin{code}
crossProd :: [[obj]] -> [obj]
crossProd xs = concat $ replicate (product $ map length (init xs)) (last xs)
\end{code}
works. Obviously, \smalltt{crossProd} is parametrized by the list \smalltt{xs} of the %%@
objects whichs 
cross product shall be generated. But again, Haskell's strong typing turns into a %%@
demerit. It would 
be nice if we could represent the resulting cross product by a list of tuples. Thus, for %%@
instance, the product $\{1,2\}\times\{3,4\}$ could be represented by the list\\
\centerline{ \smalltt{[(1,2),(1,4),(2,3),(2,4)]}.} 
But, as shown above, we aim at an instance of a Goguen category (\smalltt{Gog}) where all %%@
objects have to be of the same type. Thus, the lists \smalltt{[1,2]}, \smalltt{[3,4]} and 
\smalltt{[(1,2),(1,4),(2,3),(2,4)]} cannot be objects of the same Goguen category. The %%@
escape we use 
is to simply replicate the elements of the last object of \smalltt{xs} $n$ times whereas %%@
$n$ is the 
product of the lengths of all objects out of \smalltt{xs} except the last one. The %%@
resulting list then is isomorphic to the cross product. In our example we get %%@
\smalltt{[3,4,3,4]}. \\
The interpretability suffers from this construction. Hence, we provide the function
\begin{code}
getLabel :: Int -> [[obj]] -> [Int]
getLabel i srcs = getLabel' i $ map length srcs
  where
    getLabel' i [l]    = [i]
    getLabel' i (l:ls) = div i (product ls) : getLabel' (mod i $ product ls) ls
\end{code}
which delivers the interpretation of a position in the product of several lists. To do so %%@
it has to 
get the position in question (\smalltt{i}) and the sources \smalltt{srcs} from which the %%@
cross 
product was created. The result then is a list of positions --- one for each list out of %%@
\smalltt{srcs}. In our example, typing \smalltt{getLabel 2 [[1,2],[3,4]]} results in the %%@
output \smalltt{[1,0]}. Notice that counting the position starts with zero. Notice %%@
furthermore, that \smalltt{getLabel} presupposes that the cross product was created {\em %%@
right associative} (as shown with \smalltt{crossProd}) whereas the arrangement of the %%@
elements of 
\smalltt{srcs} is decisive. This convention comes due to the fact that it allows a %%@
relatively 
comfortable computation of the (crisp) projections, which is done by the following %%@
function. 
\begin{code}
projections :: Gog (Loos e) [obj] (FRel e obj obj) -> 
               Lat e -> [[obj]] -> [obj] -> [FRel e obj obj]
projections g e srcs trg = projs' 1 srcs
  where
    projs' _ []     = []
    projs' n (s:ss) = 
       let l1 = length s - 1
           l2 = product (map length (ss)) - 1
       in fUpd (gog_bottom g trg s) [( (k*(l1+1)*(l2+1)+j*(l2+1)+i,j)
                                        ,lat_topEl e) | k<-[0..n-1],
                                                        i<-[0..l2],
                                                        j<-[0..l1]]
          : projs' (n*(l1+1)) ss
\end{code}
Consider, for example, the three sets $\{1,2,3\}$, $\{4,5\}$ and $\{6,7\}$. The %%@
relational product of 
$\{1,2,3\}\times(\{5,6\}\times\{8,9\})$ with the convention above then is computed by the %%@
projections 
shown in Figure \ref{TabProj}.
\begin{figure}[h]
\centering
\begin{tabular}{|l||lll||ll||ll|}
%Row: 1
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf src/trg}\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf 1}\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf 2}\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf 3}\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf 5}\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf 6}\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf 8}\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf 9}\hfil}} \\
\hline
%Row: 2
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (1,5,8) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} \\
%Row: 3
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (1,5,9) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} \\
%Row: 4
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (1,6,8) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} \\
%Row: 5
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (1,6,9) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} \\
%Row: 6
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (2,5,8) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} \\
%Row: 7
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (2,5,9) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} \\
%Row: 8
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (2,6,8) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} \\
%Row: 9
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (2,6,9) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} \\
%Row: 10
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (3,5,8) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} \\
%Row: 11
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (3,5,9) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} \\
%Row: 12
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (3,6,8) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} \\
%Row: 13
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf (3,6,9) }\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $0$\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil $1$\hfil}} \\
%Row: 14
\cline{1-8}
\vbox to1,48ex{\vspace{1pt}\vfil\hbox to11,20ex{\hfil {\bf proj.} \hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil \hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf $\rho_{1}$}\hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil \hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf $\rho_{2}$}}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil \hfil}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil {\bf $\rho_{3}$}}} & 
\vbox to1,88ex{\vspace{1pt}\vfil\hbox to7,20ex{\hfil \hfil}} \\
\cline{1-8}
\end{tabular}
\caption{\label{TabProj} The projections for $\{1,2,3\}\times (\{5,6\}\times \{8,9\})$}
\end{figure}\\
With this, $1$ and $0$ are the greatest resp.\ least element of the underlying entry %%@
lattice of the 
$\SL$-fuzzy relations. One can see the periodicity of the $1$-entries which makes the %%@
computation 
quite comfortable. In $\rho_{1}$ we have period $4$, i.e., we have four $1$-entries below %%@
each other 
in every column. This comes due to the fact that $4=\mid\{5,6\}\mid %%@
\cdot\mid\{8,9\}\mid$.
Analogously, we have period $2=\mid\{8,9\}\mid$ with $\rho_{2}$ and so on. With these %%@
remarks it 
should be clear how \smalltt{projections} works. It takes the underlying Goguen category %%@
\smalltt{g}, 
the underlying entry lattice \smalltt{e} to have access to its least resp.\ greatest %%@
element, and the 
objects \smalltt{srcs} from which the cross product \smalltt{trg} was created. The %%@
computation of the 
projections then is straightforward. Notice that all terms are computed on the %%@
abstraction level of 
Goguen categories since \smalltt{projections} shall support the determination of the core %%@
of a 
controller. 

The function
\begin{code}
injections :: Gog (Loos e) [obj] (FRel e obj obj) -> 
              Lat e -> [[obj]] -> [obj] -> [FRel e obj obj]
injections g e srcs trg = inj' srcs 0
  where
    inj'[]     _ = []
    inj'(s:ss) n = fUpd (gog_bottom g s trg) 
                        [((i,j),lat_topEl e) | i <- [0..length s-1],
                                               j <- [i..length trg-1], i+n==j]
                   : inj' ss (n+length s)
\end{code}
delivers the counterpart of \smalltt{projections} by computing the (crisp) injections for %%@
a direct 
sum \smalltt{trg} that was created out of the objects \smalltt{srcs}. Again, the %%@
computation is 
straightforward whereas the arrangement of \smalltt{srcs} is decisive.

Now, we are ready to compute the core of a controller. To do so, we make use of the %%@
underlying equations (cf.\ Section \ref{SubSecOpsOnFCs}).
\begin{code}
core :: (Eq obj, Eq e) => FuzContr e obj -> FRel e obj obj
core fc@(SimpleContr (Header _ _ _ _ _ _) agg rs lEIn lEOut) =
  let g = contrCat fc
  in foldl (gog_derOp g "Meet" (opLoos agg)) 
           (gog_bottom g (contrDom fc) (contrRan fc)) $ 
             map (\(intFun,(x,y)) -> gog_derOp g "Comp" (opLoos intFun)
                                    (gog_converse g x) y) 
                 (sortEnt lEIn lEOut rs)
  where
    getMor les s = let ents = filter ((== s).entLabel) les 
                   in if null ents then error "No such entity !"
                      else head ents
    sortEnt _    _    []            = []
    sortEnt lesI lesO ((iF,(s,[])):ss)   = sortEnt lesI lesO ss
    sortEnt lesI lesO ((iF,(s,t:ts)):ss) = 
           (iF, (entRel $ getMor lesI s,
                 entRel $ getMor lesO t)) : sortEnt lesI lesO ((iF,(s,ts)):ss)
core fc@(SumContr _ sS tS derJoin subctrs) =
  let
    g        = contrCat fc;  
    doms     = map contrDom subctrs;  dom = concat doms  
    rans     = map contrRan subctrs;  ran = concat rans
    subCores = case (sS,tS) of
                 (True,True)   -> injR [] rans . injL [] doms
                 (True,False)  -> injL [] doms
                 (False,True)  -> injR [] rans
                 (False,False) -> id
  in 
    foldl1 (gog_derOp g "Meet" (opLoos derJoin)) $ subCores $ map core subctrs
  where 
    injL us [l]    (x:xs) = [fExtSrc us x []]
    injL us (l:ls) (x:xs) = fExtSrc us x ls : injL (us++[l]) ls xs
    injR us [l]    (x:xs) = [fExtTrg us x []]
    injR us (l:ls) (x:xs) = fExtTrg us x ls : injR (us++[l]) ls xs
    
core fc@(ProdContr h sP tP derMeet subctrs) =
  let
    g        = contrCat fc
    compG    = gog_comp g
    doms     = map contrDom subctrs;  dom = crossProd doms
    rans     = map contrRan subctrs;  ran = crossProd rans
    projL xs = zipWith compG (projections g (entryLat h) doms dom)xs
    projR xs = zipWith (\r p -> compG r $ gog_converse g p) xs $
                       projections g (entryLat h) rans ran
    subCores = case (sP,tP) of
                 (True,True)   -> projR . projL
                 (True,False)  -> projL
                 (False,True)  -> projR
                 (False,False) -> id
  in foldl1 (gog_derOp g "Meet" (opLoos derMeet)) $ subCores $ map core subctrs    
\end{code}
We again have to differentiate between the three cases \smalltt{SimpleContr}, %%@
\smalltt{SumContr} and \smalltt{ProdContr}. The computation for simple controllers is %%@
straightforward by reading out the rulebase and interpreting the corresponding rules. %%@
With sum and product controllers we have to treat the four different cases arising from %%@
the setting of the flags \smalltt{sS} / \smalltt{tS} resp.\ \smalltt{sP} / \smalltt{tP}. %%@
According to their state, the cores of the subcontrollers have to be extended by the %%@
(automatically generated) injections and projections, respectively. We want to mention a %%@
difference between the computations for sum controllers and product controllers. From %%@
Lemma \ref{LemInjJoin} we know that a term %%@
$\iota_{1}^{\conv}R_{1}\sqcup_{\oplus}\iota_{2}^{\conv}R_{2}$ is equivalent to %%@
$\iota_{1}^{\conv}R_{1}\sqcup\iota_{2}^{\conv}R_{2}$ for relations $R_{1}:A_{1}\lora B$, %%@
$R_{2}:A_{2}\lora B$ and the induced crisp injections $\iota_{i}:A_{i}\lora A_{1}+A_{2}$ %%@
if $\sqcup_{\oplus}$ is a meet-based derived operation with neutral element $\Bot$. Thus, %%@
applying the injections in this term corresponds to simply extending the matrices $R_{1}$ %%@
resp.\ $R_{2}$ by zero rows such that the resulting matrices $R_{1}'$ and $R_{2}'$ have %%@
source $A_{1}+A_{2}$ and target $B$. This means that the operations \smalltt{fExtSrc} and %%@
\smalltt{fExtTrg} of the \smalltt{LFuzzyRel} module are applicable. An analogous %%@
computation for the relational product is obviously not to expect. 
\subsubsection* { Combinators for fuzzy controllers }
After the basic routines on fuzzy controllers, we now want to provide suitable %%@
combinators to construct them step by step. Furthermore, we then have the possibility to %%@
construct new fuzzy controllers in a component-free manner. Building up a fuzzy %%@
controller always starts with the empty controller.
\begin{code}
emptyContr l unit eLat = let emptyLoos = Loos { loos_lat = eLat
                                               ,loos_op  = const $ id
                                               ,loos_e   = lat_botEl eLat
                                               ,loos_z   = lat_botEl eLat}
                             emptyDerOp = DerOp "" emptyLoos True True
                             emptyFunc  = UnFunc "" id
                             emptyBFunc = BinFunc "" (const $ id)
                         in if l=="" then error ("Empty label!")
                            else SimpleContr (Header l unit eLat  emptyFunc
                                                       emptyBFunc emptyFunc) 
                                             emptyDerOp [] [] []
\end{code}
Obviously, \smalltt{emptyContr} only takes a minimal set of parameters. The remaining %%@
part can be set to default values. Notice that a controller of type \smalltt{SumContr} or %%@
\smalltt{ProdContr} with an empty \smalltt{subContrs} list is seen to be invalid. %%@
Furthermore, the empty label is not allowed for a controller. The reason for this becomes %%@
clear later on.

An empty controller can be updated by new rules and linguistic entities, respectively. %%@
Furthermore, the corresponding parts of the header can be set. This is covered by the %%@
following routines.
\begin{code}
hUpdFuzz f   (Header l u e _ ap d) = Header l u e f ap d
hUpdAppl ap  (Header l u e f _ d)  = Header l u e f ap d
hUpdDefuzz d (Header l u e f ap _) = Header l u e f ap d
 
contrUpdFuzz f (SimpleContr h ag rs lEIn lEOut) = 
   SimpleContr (hUpdFuzz f h) ag rs lEIn lEOut
contrUpdFuzz f (SumContr h sS tS dJ sC) =
   SumContr (hUpdFuzz f h) sS tS dJ sC
contrUpdFuzz f (ProdContr h sP tP dM sC) =
   ProdContr (hUpdFuzz f h) sP tP dM sC

contrUpdAppl ap (SimpleContr h ag rs lEIn lEOut) =
   SimpleContr (hUpdAppl ap h) ag rs lEIn lEOut
contrUpdAppl ap (SumContr h sS tS dJ sC) =
   SumContr (hUpdAppl ap h) sS tS dJ sC
contrUpdAppl ap (ProdContr h sP tP dM sC) =
   ProdContr (hUpdAppl ap h) sP tP dM sC

contrUpdAggr ag (SimpleContr h _ rs lEIn lEOut) =
    SimpleContr h ag rs lEIn lEOut

contrUpdDefuzz d (SimpleContr h ag rs lEIn lEOut) =
   SimpleContr (hUpdDefuzz d h) ag rs lEIn lEOut
contrUpdDefuzz d (SumContr h sS tS dJ sC) =
   SumContr (hUpdDefuzz d h) sS tS dJ sC
contrUpdDefuzz d (ProdContr h sP tP dM sC) =
   ProdContr (hUpdDefuzz d h) sP tP dM sC

contrUpdRuleIn, contrUpdRuleOut :: 
    String -> [Rule e] -> FuzContr e obj -> FuzContr e obj
contrUpdRuleIn cLabel xs fc@(SimpleContr (Header l u e f ap d) ag rs lI lO) = 
    if cLabel == "" || l == cLabel then
      SimpleContr (Header l u e f ap d) ag (xs++rs) lI lO
    else fc
contrUpdRuleIn cLabel xs (SumContr  (Header l u e f ap d) dJ sS tS subctrs) =
    if cLabel == "" || l == cLabel then 
      error ("Label \""++cLabel++"\" specifies a sum controller"++
             " or is invalid !")
    else SumContr (Header l u e f ap d) dJ sS tS $ 
                  map (contrUpdRuleIn cLabel xs) subctrs
contrUpdRuleIn cLabel xs (ProdContr (Header l u e f ap d) dM sP tP subctrs) = 
    if cLabel == "" || l == cLabel then 
      error ("Label \""++cLabel++"\" specifies a product controller"++
             " or is invalid !")
    else ProdContr (Header l u e f ap d) dM sP tP $ 
         map (contrUpdRuleIn cLabel xs) subctrs

contrUpdRuleOut l xs = 
    contrUpdRuleIn l (concat $ map (\(iF,(t,ss)) -> [(iF,(s,[t])) | s<-ss]) xs)

contrUpdEntIn, contrUpdEntOut :: (Eq obj) => 
      String -> [(DerOp e,
                 (LingEntity (FRel e obj obj),[String]))] -> 
      FuzContr e obj -> FuzContr e obj
contrUpdEntIn cLabel lEs fc@(SimpleContr (Header l u e f ap d) ag rs lI lO) = 
    if cLabel == "" || l == cLabel then
      SimpleContr (Header l u e f ap d) ag
                  (map (\(iF,(e,ts)) -> (iF, (entLabel e,ts))) lEs ++ rs)
                  (map (fst.snd) lEs ++ lI)
                  lO
    else fc
contrUpdEntIn cLabel lEs (SumContr (Header l u e f ap d) dJ sS tS subctrs) =
    if cLabel == "" || l == cLabel then
      error ("Label \""++cLabel++"\" specifies a sum controller"++
             " or is invalid !")    
    else SumContr (Header l u e f ap d) dJ sS tS $
              map (contrUpdEntIn cLabel lEs) subctrs
contrUpdEntIn cLabel lEs (ProdContr (Header l u e f ap d) dM sP tP subctrs) =
    if cLabel == "" || l == cLabel then
      error ("Label \""++cLabel++"\" specifies a product controller"++
             " or is invalid !")    
    else ProdContr (Header l u e f ap d) dM sP tP $ 
              map (contrUpdEntIn cLabel lEs) subctrs

contrUpdEntOut cLabel lEs fc@ (SimpleContr (Header l u e f ap d) ag rs lI lO) = 
    if cLabel == "" || l == cLabel then
      SimpleContr (Header l u e f ap d) ag
                  (concat (map (\(iF,(e,ts)) -> [(iF,(t,[entLabel e]))
                                                 | t<-ts]) lEs) ++ rs)
                  lI
                  (map (fst.snd) lEs ++ lO)
    else fc
contrUpdEntOut cLabel lEs (SumContr (Header l u e f ap d) dJ sS tS subctrs) =
    if cLabel == "" || l == cLabel then
      error ("Label \""++cLabel++"\" specifies a sum controller"++
             " or is invalid !")    
    else SumContr (Header l u e f ap d) dJ sS tS $ 
              map (contrUpdEntOut cLabel lEs) subctrs
contrUpdEntOut cLabel lEs (ProdContr (Header l u e f ap d) dM sP tP subctrs) =
    if cLabel == "" || l == cLabel then
      error ("Label \""++cLabel++"\" specifies a product controller"++
             " or is invalid !")    
    else ProdContr (Header l u e f ap d) dM sP tP $ 
              map (contrUpdEntOut cLabel lEs) subctrs
\end{code}
The functions for the header should be clear. For updating the rulebase, we have %%@
\linebreak \smalltt{contrUpdRuleIn} resp.\ \smalltt{contrUpdRuleOut}. For the case that %%@
the rulebase of a subcontroller within a sum or product controller is to update, we %%@
introduce the parameter \smalltt{cLabel}. As shown here, the \smalltt{SumContr} and %%@
\smalltt{ProdContr} parts of every function shown above are recursive. The operation in %%@
question is then applied to {\em every} subcontroller having the name \smalltt{cLabel}. %%@
If \smalltt{cLabel} specifies a sum or product controller, the functions above fail with %%@
an error message. Notice again, that we demand the controller labels to be unique.\\
For simple controllers \smalltt{cLabel} plays no role and, hence, the user can set it to %%@
\smalltt{""}. Here we see the reason why controllers with an empty label are not %%@
allowed.\\
Back to the functions for updating rules, the parameter \smalltt{xs} specifies the new %%@
rules carrying the list of new target entities a source entity shall be connected with. %%@
Notice that we have a difference here between updating rules for input entities and %%@
updating them for output entities. With \smalltt{contrUpdRuleIn} the source entity is an %%@
input entity and the targets are output entities. This exactly corresponds to our %%@
interpretation of the data type \smalltt{Rule}. But, with \smalltt{contrUpdRuleOut} the %%@
source is an output entity. Hence, we have to modify the list in the way shown above such %%@
that we can use \smalltt{contrUpdRuleIn} to compute the resulting controller.\\
The functions for updating linguistic entities work analogously. The only interesting %%@
thing is the structure of the new parameter \smalltt{lEs}. It consists of a list of %%@
tuples each carrying the new linguistic entity and a list of entities it shall be %%@
connected with in the rulebase. Hence, we are able to update the rulebase simultaneously %%@
without having to perform a separate function call. 

All these operations above are still somehow componentwise. The next constructions are %%@
based on ready created fuzzy controllers. The product of two controllers can be build by %%@
the function
\begin{code}
contrProd, contrSum :: (Eq e, Eq obj) => Bool -> Bool -> FuzContr e obj -> 
                                        FuzContr e obj ->  DerOp e -> FuzContr e obj
contrProd sP tP c1 c2 derMeet  = 
    let dummy         = ProdContr (header c1) sP tP derMeet [c1,c2]
        eLat          = entryLat $ header c1;  g     = contrCat dummy
        compG         = gog_comp g;            convG = gog_converse g
        dom1          = contrDom c1;           dom2  = contrDom c2
        ran1          = contrRan c1;           ran2  = contrRan c2
        pI            = projections g eLat [dom1,dom2] $ crossProd [dom1,dom2]
        pO            = projections g eLat [ran1,ran2] $ crossProd [ran1,ran2]
        (newF,newApp) = 
          let f1  = fuzz (header c1); f2  = fuzz (header c2)
              ap1 = appl (header c1); ap2 = appl (header c1)
              a1  = binFunc ap1;      a2  = binFunc ap2
          in
           if not sP then (f1,ap1)
           else 
            (UnFunc (unLabel f1++"**"++unLabel f2)
              (\x -> foldl1 (gog_derOp g "Meet" (opLoos derMeet)) $ 
                     zipWith (\p f -> compG (f $ compG x p) $ convG p)
                     pI [unFunc f1,unFunc f2]),
             BinFunc (binLabel ap1++"**"++binLabel ap2)
              (\x y -> gog_derOp g "Meet" (opLoos derMeet)
                       (compG (a1 (compG x (head pI)) $ 
                                  (compG (compG (convG $ head pI) y) $ head pO))
                              $ convG (head pO) )
                       (compG (a2 (compG x (last pI)) $ 
                                  (compG (compG (convG $ last pI) y) $ last pO))
                              $ convG (last pO) ) ) )
        newDF         = 
          let df1 = defuzz (header c1); df2 = defuzz (header c2) in
            if not tP then df1
            else UnFunc (unLabel df1++"**"++unLabel df2)
                        (\x -> foldl1 (gog_derOp g "Meet" (opLoos derMeet)) $ 
                               zipWith (\p f -> compG (f $ compG x p) $ convG p)
                               pO [unFunc df1,unFunc df2])
    in  ProdContr (Header
                     (contrLabel (header c1)++"**"++contrLabel (header c2))
                     (unitObj $ header c1) (entryLat $ header c1)
                     newF newApp newDF )
                  sP tP
                  derMeet
                  [c1,c2]
\end{code}
which is straightforward due to the explanations in Section \ref{SubSecOpsOnFCs}. The %%@
only things that have to be computed are the fuzzification, application and %%@
defuzzification functions in conformance with the flags \smalltt{sP} and \smalltt{tP}. %%@
This is done by \smalltt{newF}, \smalltt{newAppl} and \smalltt{newDF}, respectively. The %%@
label of the resulting controller is simply created by connecting the labels of %%@
\smalltt{c1} and \smalltt{c2} with an additional \smalltt{**}. Notice that we do not have %%@
to differentiate between the three variants of \smalltt{FuzContr} since we only need %%@
parameters of the header of \smalltt{c1} and \smalltt{c2}. But, notice that the function %%@
is {\em not} associative in the sense that the resulting data structures are equal. But, %%@
for the mathematical interpretation (the core) this plays no role since the relational %%@
product is associative. Furthermore, we use the labels of two controllers as the equality %%@
criteria in our \smalltt{Eq} instance. As one can see, the computation of the label of %%@
the resulting controller in \smalltt{contrProd} is associative. Notice that the operation %%@
above preserves only the properties both \smalltt{c1} and \smalltt{c2} have, i.e., a rule %%@
of the product controller can only fire if this rule would cause \smalltt{c1} and %%@
\smalltt{c2} to fire. This is immediately seen by the fact that the loos %%@
\smalltt{derMeet} is restricted to have $\iden$ as neutral element and, hence, $\Bot$ as %%@
zero.

The next construction is the relational sum of two controllers.
\begin{code}
contrSum sS tS c1 c2 derJoin  =
    let dummy         = SumContr (header c1) sS tS derJoin [c1,c2]
        eLat          = entryLat $ header c1;  g     = contrCat dummy
        compG         = gog_comp g;            convG = gog_converse g
        dom1          = contrDom c1;           dom2  = contrDom c2
        ran1          = contrRan c1;           ran2  = contrRan c2
        iI            = injections g eLat [dom1,dom2] $ dom1++dom2
        iO            = injections g eLat [ran1,ran2] $ ran1++ran2
        (newF,newApp) = 
          let f1  = fuzz (header c1); f2  = fuzz (header c2)
              ap1 = appl (header c1); ap2 = appl (header c1)
              a1  = binFunc ap1;      a2  = binFunc ap2
          in
           if not sS then (f1,ap1)
           else 
            (UnFunc (unLabel f1++"++"++unLabel f2)
               (\x -> foldl1 (gog_derOp g "Meet" (opLoos derJoin)) $ 
                      zipWith (\i f -> compG (f $ compG x (convG i)) i)
                      iI [unFunc f1,unFunc f2]),
             BinFunc (binLabel ap1++"++"++binLabel ap2)
               (\x y -> gog_derOp g "Meet" (opLoos derJoin)
                        (compG (a1 (compG x $ convG (head iI)) $ 
                                   (compG (compG (head iI) y) $ convG (head iO)))
                               $ head iO )
                        (compG (a2 (compG x $ convG (last iI)) $ 
                                   (compG (compG (last iI) y) $ convG (last iO)))
                               $ last iO ) ) )
        newDF         = 
          let df1 = defuzz (header c1); df2 = defuzz (header c2) in
          if not tS then df1
          else  UnFunc (unLabel df1++"++"++unLabel df2)
                       (\x -> foldl1 (gog_derOp g "Meet" (opLoos derJoin)) $ 
                              zipWith (\i f -> compG (f $ compG x (convG i)) i)
                              iO [unFunc df1,unFunc df2])
    in  SumContr (Header 
                    (contrLabel (header c1)++"++"++contrLabel (header c2))
                    (unitObj $ header c1) (entryLat $ header c1)
                    newF newApp newDF )
                  sS tS
                  derJoin
                  [c1,c2]
\end{code}
The implementation is analogous and a direct realization of the underlying operation. The %%@
following two functions mark the derived meet resp.\ derived join operation on two fuzzy %%@
controllers as special cases for the functions above.
\begin{code}
contrJoin, contrMeet :: (Eq e, Eq obj) => FuzContr e obj -> FuzContr e obj -> 
                                          DerOp e -> FuzContr e obj
contrJoin = contrSum  False False
contrMeet = contrProd False False
\end{code}
Now, we want to switch to \smalltt{contrMelt}. The main motivation for this function is %%@
the reduction of the underlying data structure \smalltt{FuzContr} and a gain in %%@
efficiency for several computations (e.g., \smalltt{contrSum}). The resulting controller %%@
shall be of type \smalltt{SimpleContr} and contain at maximum one kind of linguistic %%@
entities for each input and output, i.e., all input/output entities shall have the same %%@
target. But, notice that the history of creation of the underlying controller is %%@
automatically deleted when using \smalltt{contrMelt} and cannot be reconstructed. We %%@
start with sum controllers.
\begin{code}
contrMelt :: (Eq e, Eq obj) => FuzContr e obj -> FuzContr e obj
contrMelt fc@(SumContr h sS tS derJoin subctrs) =
  let
    subms  = map contrMelt subctrs;   ag    = aggr $ head subms
    doms   = map contrDom subms;      rans  = map contrRan subms
    lEIn   = map lingEntIn subms;     lEOut = map lingEntOut subms
    rs     = concat $ map rules subms
    extLab = case (sS,tS) of
              (True,True)  -> map (\(iF,(s,ts))->(iF,(s++"++",map (++"++") ts)))
              (True,False) -> map (\(iF,(s,ts))->(iF,(s++"++",ts)))
              (False,True) -> map (\(iF,(s,ts))->(iF,(s,map(++"++") ts)))
              (False,False)-> id
  in SimpleContr h ag
                 (extLab $ meltRs (comm ag) (idem ag) (head rs) [] $ tail rs)
                 (nub $ concat $ if sS then injR [] doms lEIn  else lEIn)
                 (nub $ concat $ if tS then injR [] rans lEOut else lEOut)
  where
    injR us [l]    (x:xs) = [map (\le -> LingEntity (entLabel le++"++") 
                                                    (fExtTrg us (entRel le)[]))x]
    injR us (l:ls) (x:xs) =  map (\le -> LingEntity (entLabel le++"++") 
                                                    (fExtTrg us (entRel le)ls)) x
                             : injR (us++[l]) ls xs
    meltRs _ _ x               []     []  = [x]
    meltRs c i x               (r:rs) []  = x:meltRs c i r [] rs
    meltRs c i x@(iF,(l,trgs)) rs     (t@(iF2,(l2,trgs2)):ts)= 
         if (l==l2) && (opLabel iF==opLabel iF2) && c then 
             meltRs c i (iF,(l,if i then nub $ trgs++trgs2 
                               else trgs++trgs2))          rs ts
         else meltRs c i x (t:rs) ts
\end{code}
The header need not be computed again. This has already been done with the creation of %%@
the data structure using \smalltt{contrSum}. The aggregation function is delivered by the %%@
melted subcontrollers (\smalltt{subms}). Notice that we choose the first element of this %%@
list. Indeed, we could choose an arbitrary element since we implicitly presuppose that %%@
the subcontrollers are combinable.\\ 
The computation of the new linguistic entities and rulebase is a straightforward %%@
realization of Definitions \ref{DefnDerSum} and \ref{DefnDerSumMod}. Notice that the %%@
resulting input and output entities get new labels if they are extended to new targets. %%@
For the computation of the new rulebase we explicitly consider whether the aggregation %%@
function is commutative (\smalltt{c}) or idempotent (\smalltt{i}). Idempotency implies %%@
that we can omit duplicate output entities within a rule, and commutativity allows, %%@
presupposed equal interpretation functions, to put two rules together, which have the %%@
same input entity. Finally, we again want to mention that this construction does not %%@
affect the input-output behavior of the resulting controller (cf.\ Theorem %%@
\ref{ThmInOutSum}).\\
Melting two product controllers is done analogously.
\begin{code}
contrMelt fc@(ProdContr h sP tP derMeet subctrs) =
  let g        = contrCat fc;          subms = map contrMelt subctrs
      lEIn     = map lingEntIn subms;  lEOut = map lingEntOut subms
      compG    = gog_comp g;           convG = gog_converse g
      concProj = zipWith (\p (LingEntity l r) -> 
                             (LingEntity l (compG r $ convG p)))
      extLab  = case (sP,tP) of
                (True,True)   -> map (\(s,ts) -> (s++"&&",map (++"&&") ts))
                (True,False)  -> map (\(s,ts) -> (s++"&&",ts))
                (False,True)  -> map (\(s,ts) -> (s,map(++"&&") ts))
                (False,False) -> id
  in SimpleContr h (aggr $ head subms)
                 (newRules $ map rules subms)
                 (nub $ prodEnts  concProj g
                                  lEIn (map contrDom subms) sP)
                 (nub $ prodEnts  concProj g
                                  lEOut (map contrRan subms) tP)
  where
     newRules []         = []
     newRules [rs]       = rs
     newRules (rs:ruls)  = [(iF,(l ++ (if sP then "**" else "&&") ++ l2,
                                 [n++ (if tP then "**" else "&&") ++ n2 
                                  | n<-ls,n2<-ls2]))
                             |(iF,(l,ls)) <- rs, (iF2,(l2,ls2)) <- newRules ruls]
     prodEnts _ _      [] _  _    = error ("Empty entity list while "++
                                           "computing the relational products!")
     prodEnts f g ents trgs cP = 
         let es          = prodEnts' ents
             cp          = crossProd trgs
             pr          = projections g (entryLat h) trgs cp
             (f',t,conc) = if cP then (f pr,cp,"**")
                           else (id,head trgs,"&&")
         in map (\en -> LingEntity (tail $ tail $ foldl (\x y -> x ++ conc ++
                                                             entLabel y) "" en)  
                                   (foldl (\x y -> gog_derOp g "Meet" 
                                                     (opLoos derMeet) x (entRel y))
                                                     (gog_top g (unitObj h) t)
                                                     $ f' en)  ) es
     prodEnts' []        = error ("Empty entity list while computing " ++
                                  "the relational products!")
     prodEnts' [es]      = map (:[]) es
     prodEnts' (es:ents) = let en = prodEnts' ents
                                in  concat [[e1:es2 | es2<-en] | e1<-es]
\end{code}
The key functionality is established by \smalltt{prodEnts} and \smalltt{prodEnts'}, %%@
respectively. The second function creates all pairs of linguistic entities arising from %%@
the melted subcontrollers. These pairs are represented by lists. The lists are the base %%@
for \smalltt{prodEnts} to generate the projections, if necessary, and then compute the %%@
resulting linguistic entities (cf.\ Definitions \ref{DefnDerProd}, \ref{DefnDerProdMod}). %%@
The entities of the melted controller also get new labels. If they are extended using the %%@
projections (i.e., \smalltt{sP} or \smalltt{tP} are \smalltt{True}), they are %%@
concatenated by \smalltt{**}. Otherwise, we use \smalltt{\&\&}. This labeling is a good %%@
indicator if one wants to list the resulting controller using the \smalltt{show} %%@
function. \\
With product controllers we have to remember that the input-output behavior of the melted %%@
controller is affected. This immediately follows from Theorem \ref{ThmInOutProd}. But, we %%@
have in every case that calling\\ 
\centerline{ \smalltt{core \$ contrProd sProd tProd c1 c2 derM}}
and \\
\centerline{ \smalltt{core (contrMelt \$ contrProd sProd tProd c1 c2 derM)}}
deliver the same result for given parameters \smalltt{sProd}, \smalltt{tProd}, %%@
\smalltt{c1}, \smalltt{c2} and \smalltt{derM}. The same is true for \smalltt{contrSum}.

The standard case of \smalltt{contrMelt} (for simple controllers) is trivial. 
\begin{code}
contrMelt f = f
\end{code}
Finally, we want to introduce predefined functions to infer the output for a given %%@
controller and a given input. 
\begin{code}
fuzzify fc x   = unFunc (fuzz $ header fc) x
apply fc x     = binFunc (appl $ header fc) x $ core fc
defuzzify fc y = unFunc (defuzz $ header fc) y
infer fc       = defuzzify fc . apply fc . fuzzify fc
\end{code}
They can be used quite intuitively.

With these combinators a convenient use of the module should be possible. The user now %%@
has the chance to manually and (in wide areas) automatically construct and manipulate %%@
fuzzy controllers.%%@
