\section{ Example controller }
In this section we want to provide the module
\begin{code}
module FContrTest where

import FContr
import Lattice
import LFuzzyRel
import LFuzzyRelLattices
import List (elemIndex)
\end{code}
in which we develop a little example controller to show how to use \smalltt{FContr}. With %%@
the example we do not aim at a sophisticated controller, but try to involve as many %%@
operations as possible.

Suppose we have the crossroad shown in Figure \ref{FigModContr} where the traffic is %%@
regulated by two traffic lights ``L1'' and ``L2''.
\begin{figure}[h]
\begin{center}
\begin{picture}(178,166)
\begin{tiny}
\thinlines
\drawpath{4.0}{98.0}{76.0}{98.0}
\drawpath{76.0}{98.0}{76.0}{162.0}
\drawpath{4.0}{68.0}{76.0}{68.0}
\drawpath{76.0}{68.0}{76.0}{4.0}
\drawpath{108.0}{4.0}{108.0}{68.0}
\drawpath{108.0}{68.0}{174.0}{68.0}
\drawpath{174.0}{68.0}{174.0}{68.0}
\drawpath{174.0}{98.0}{108.0}{98.0}
\drawpath{108.0}{98.0}{108.0}{162.0}
\drawframebox{120.0}{105.0}{12.0}{10.0}{L1}
\drawframebox{65.0}{106.0}{10.0}{12.0}{L2}
\drawdotline{4.0}{82.0}{4.0}{82.0}
\drawdotline{4.0}{82.0}{6.0}{82.0}
\drawdotline{6.0}{82.0}{172.0}{82.0}
\drawdotline{92.0}{160.0}{92.0}{4.0}
\drawvector{144.0}{90.0}{32.0}{-1}{0}
\drawvector{40.0}{76.0}{32.0}{1}{0}
\drawvector{84.0}{132.0}{32.0}{0}{-1}
\drawvector{100.0}{34.0}{32.0}{0}{1}
\drawpath{126.0}{110.0}{134.0}{104.0}
\drawpath{126.0}{100.0}{132.0}{104.0}
\drawpath{60.0}{112.0}{64.0}{118.0}
\drawpath{70.0}{112.0}{64.0}{118.0}
\drawcenteredtext{122.0}{76.0}{Street 1}
\drawcenteredtext{122.0}{126.0}{Street 2}
\end{tiny}
\end{picture}
\caption{\label{FigModContr} A crossroad controlled by a fuzzy controller}
\end{center} 
\end{figure}
 The signal flow of both traffic lights shall be dependent from the amount of traffic on %%@
both streets. Our aim is to develop a controller steering both ``L1'' and ``L2''. This %%@
can be achieved by creating two separate controllers and then applying the cross product %%@
operation to them. This corresponds to a kind of ``concurrent'' modeling.

A first simple approach to control a single traffic light is the following. We want to %%@
make the controller dependent from the amount of traffic on the two streets. Hence, we %%@
have to get an input from two sensors each measuring the traffic of its street. We denote %%@
the output domain of a sensor by $A$ and, thus, have the input space $A\times A$ for the %%@
controllers. We provide the linguistic input entities
\begin{tabbing}
\qquad \= $LT : I\lora A\times A$   \= - \= low traffic, \\
       \> $MT$                      \> - \> medium traffic, \\
       \> $HT$                      \> - \> high traffic.
\end{tabbing}
The output entities are the following :
\begin{tabbing}
\qquad \= $TR : I\lora B$ \= - \= turn red, \\
       \> $S$             \> - \> stay unchanged, \\
       \> $TG$            \> - \> turn green.
\end{tabbing}
They obviously directly correspond to the possible actions a traffic light can do. The %%@
range $B$ includes the possible states of a traffic light. Normally, we would have  $red, %%@
yellow$ and $green$. But, $yellow$ is a time-dependent (and not traffic-dependent) state. %%@
Since we do not have modeled time in our approach, we omit this state so that we have %%@
$B:=\{red, green\}$.

Now, the question arises which entry lattice to choose. An entry in the input entities %%@
shall express to which degree the traffic on street one and two, respectively, is low, %%@
medium or high. Hence, we choose $[0,1]\times[0,1]$ as the underlying entry lattice.

At this point we have enough information to model the single controllers ``L1'' and %%@
``L2''. First, we provide the entry lattice. It is clear that we have to use an %%@
approximation of $[0,1]\times[0,1]$. We use a granularity of $\frac{1}{100}$.
\begin{code}
gran = (100::Int)
eLat :: Lat (Int,Int)
eLat = let els = [ (x,y) | x<-[0..gran], y<-[0..gran]]
       in Lat{lat_poSet = PoSet { poSet_isElem   = flip elem $ els
                                 ,poSet_elements = els
                                 ,poSet_lEq      = \(x1,y1) (x2,y2)->x1<=x2 && 
                                                                     y1<=y2 }
             ,lat_sup     = \(x1,y1) (x2,y2) -> (max x1 x2,max y1 y2)
             ,lat_inf     = \(x1,y1) (x2,y2) -> (min x1 x2,min y1 y2)
             ,lat_botEl   = (0,0)
             ,lat_topEl   = (gran,gran)
             ,lat_atomS   = [(1,0),(0,1)]
             ,lat_jIrredS = [(0,x)|x<-[1..gran]]++[(x,0)|x<-[1..gran]]
             ,lat_mIrredS = [(gran,x)|x<-[0..gran-1]]++[(x,gran)|x<-[0..gran-1]]}
\end{code}
The parameter \smalltt{gran} sets the granularity. The lattice then is instantiated quite %%@
intuitively. The only thing to mention is that we use the componentwise ordering\\ 
\centerline{$(x_{1},y_{1})\leq (x_{2},y_{2})\ :\lra\ x_{1}\leq x_{2}$ and $y_{1}\leq %%@
y_{2}$}
on $[0,1]\times[0,1]$ so that we cannot rely on the standard \smalltt{Prelude} function %%@
\smalltt{(<=)}. The definition above is consistent which is shown by the following %%@
session on Hugs or GHC.
\begin{smallverb}
FContrTest> lat_atomS eLat == atomSet (lat_lSemiLat eLat)
True
FContrTest> lat_mIrredS eLat == mIrredSet eLat 
True
FContrTest> lat_jIrredS eLat == jIrredSet eLat 
True
FContrTest> testLattice eLat []
[]
\end{smallverb}
Notice that checking both the join-irreducible and meet-irreducible elements before %%@
calling \smalltt{testLattice} is necessary since \smalltt{testLattice} automatically %%@
reduces the underlying tests to these elements (cf.\ Section \ref{SecLatModul}).

Now, we can start to model the linguistic entities. First, some parameters are %%@
introduced.
\begin{code}
unit,trafficSrc,tLightTrg,trafficAB :: [Int]
unit       = [1]
trafficSrc = [0..9]
tLightTrg  = [20,21]
trafficAB = crossProd [trafficSrc,trafficSrc]
\end{code}
With this \smalltt{unit} delivers the unit object which constitutes the source of the %%@
linguistic entities. Furthermore, \smalltt{trafficSrc} gives the range of the measuring %%@
devices for the traffic and can be interpreted as the amount of cars at the crossroad on %%@
the respective street. Hence, it corresponds to the set $A$ introduced above. The target %%@
of the output entities ($B$ from above) is provided by the parameter \smalltt{tLightTrg}. %%@
Since we have to type all lists equally, the entries have to be interpreted as $red$ %%@
(\smalltt{20}) and $green$ (\smalltt{21}), respectively. Finally, the cross product %%@
$A\times A$ is delivered by \smalltt{trafficAB}.

We want to construct our linguistic input entities using the intensifying modifier ``more %%@
or less''. This corresponds to the application $F;_{*}\Xi$ for a given fuzzy set $F$, %%@
modifier $\Xi$ and a derived operation $;_{*}$. We exemplary explain the procedure while %%@
constructing the linguistic entity $HT$. In the following we use $;$ for $;_{*}$. First, %%@
we introduce two given crisp fuzzy sets $F_{1}$ and $F_{2}$ over $A\times A$ expressing %%@
which amount of cars constitutes a high traffic on street one resp.\ two. With this we %%@
suppose that street one is the main street and, hence, shall be preferred with state {\em %%@
green}.
\begin{code}
highA = fUpd (fBot eLat [1::Int] trafficAB) [((0,80+i),(gran,0))|i<-[0..19]]
highB = fUpd (fBot eLat [1::Int] trafficAB) [((0,i*10+9),(0,gran))|i<-[0..9]]
\end{code}
Notice that \smalltt{trafficAB} was created using the \smalltt{crossProd} operation. %%@
Together with the explanations of Section \ref{SecModFContr} it then should be clear that %%@
\smalltt{highA} and \smalltt{highB} are mathematically described by\\
\[ F_{1}(1,(x,y)):=\left\{\begin{array}{ll}
                             (1,0) &,\ \mbox{if\ } x\geq 8\\
                             (0,0) &,\ \mbox{otherwise}                                     
                          \end{array}\right., \qquad
             F_{2}(1,(x,y)):=\left\{\begin{array}{ll}
                               (0,1) &,\ \mbox{if\ } y=9\\
                               (0,0) &,\ \mbox{otherwise}                                     
                             \end{array}\right. .
\]
Obviously, already eight cars on street one are seen as high traffic whereas nine cars %%@
are needed on street two. The starting fuzzy sets $F_{1}$ and $F_{2}$ now have to be %%@
modified. Again, we have to provide a granularity how sensitive the controller shall %%@
react on little changes in the traffic amount. To avoid unnecessarily many state changes %%@
with the traffic lights, a change in the traffic amount lower than two cars on one street %%@
shall not affect the behavior. This is modeled by the fuzzy relation $\Xi:A\times A\lora %%@
A\times A$
\begin{eqnarray}
\label{EqXi}
\Xi((x_{1},y_{1}),(x_{2},y_{2}))&:=(& min(1,max(0,1.2-0.2\cdot|x1-x2|)),\nonumber\\
                                &   & min(1,max(0,1.2-0.2\cdot|y1-y2|))).
\end{eqnarray}
One can see that $\iden_{A\times A}\sqsse\Xi$ holds. If we now compute $F_{1};\Xi$, the %%@
interpretation of ``high traffic'' is preserved within $F_{1}$. But, everything that is %%@
``nearly high traffic'' gets a weakened entry. Thus, we, for example, have %%@
$(F_{1};\Xi)(5,0)=(0.6,0)$. To achieve the overall fuzzy set $HT$, we finally have to %%@
compute $(F_{1}\sqcup F_{2});\Xi$. With these remarks the Haskell code should be %%@
understandable. 
\begin{code}
trAB :: [(Int,Int)]
trAB = [(x,y) | x<-trafficSrc,y<-trafficSrc]
mayBeVal (Just i) = i
mayBeVal Nothing  = 0
xi :: FRel (Int,Int) Int Int
xi = fUpd (fBot eLat trafficAB trafficAB)
      [((mayBeVal $ elemIndex (x1,y1) trAB, mayBeVal $ elemIndex (x2,y2) trAB),
        (min gran $ max 0 $ gran+20-20*abs (x1-x2),
         min gran $ max 0 $ gran+20-20*abs (y1-y2)) )
       | (x1,y1)<-trAB, (x2,y2)<-trAB]

highT   = LingEntity "HT" $ roughly xi (fJoin highA highB) 1
\end{code}
The function \smalltt{trAB} represents $A\times A$. It differs from \smalltt{trafficAB} %%@
because it consists of tuples and, hence, allows a comfortable creation of $\Xi$. %%@
Furthermore, \smalltt{getMayBeValue} is an auxiliary function which is also needed with %%@
\smalltt{xi}. The implementation of \smalltt{xi} then is a straightforward realization of %%@
Formula \ref{EqXi}. Finally, $HT$ is computed by \smalltt{highT} which makes use of the %%@
predefined function \smalltt{roughly} (cf.\ Section \ref{SecModFContr}).\\
The linguistic entities $MT$ and $LT$ for medium resp.\ low traffic are provided %%@
analogously. 
\begin{code}
upd1    = fUpd (fBot eLat unit trafficAB)
mediumA = upd1 [((0,30+i),(gran,0))|i<-[0..9]]
mediumB = upd1 [((0,i*10+4),(0,gran))|i<-[0..9]]

lowA = upd1 [((0,i),(gran,0))|i<-[0..9]]
lowB = upd1 [((0,i*10),(0,gran))|i<-[0..9]]

mediumT = LingEntity "MT" $ roughly xi (fJoin mediumA mediumB) 1
lowT    = LingEntity "LT" $ roughly xi (fJoin lowA lowB) 1
\end{code}
Now, we switch to the output entities. They are realized as follows.
\begin{code}
upd2 = fUpd (fBot eLat unit tLightTrg)
tRed   = LingEntity "TR" $ upd2 [((0,0),lat_topEl eLat)]
tGreen = LingEntity "TG" $ upd2 [((0,1),lat_topEl eLat)]
stay   = LingEntity "S"  $ upd2 [((0,0),(div gran 2,div gran 2)),
                                 ((0,1),(div gran 2,div gran 2))]
\end{code}
The entities $TR$ and $TG$, implemented by \smalltt{tRed} resp.\ \smalltt{tGreen}, are %%@
crisp whereas $S$ is not. This representation is chosen arbitrarily.

The input and output entities have to be connected within the rulebase. With the model %%@
developed so far we choose a very intuitive variant.
\begin{tabbing}
\qquad \= If $x$ is $HT$ then $y$ is $TG$\\
       \> If $x$ is $MT$ then $y$ is $S$\\
       \> If $x$ is $LT$ then $y$ is $TR$ 
\end{tabbing}
The rulebase demands to turn red when there is only low traffic on the respective street. %%@
Analogously, turning green is demanded with high traffic. Medium traffic shall cause the %%@
traffic light to stay unchanged.

Now, we construct ``L1'' and ``L2''. Throughout this example we want all rules to be %%@
interpreted by $;$. The application function of the resulting controllers also shall be %%@
the common composition operator. Furthermore, we use the identity operator as %%@
fuzzification, and join is used as aggregation function. Hence, we introduce the %%@
following abbreviations.
\begin{code}
infOp, supOp :: DerOp (Int,Int)
infOp       = DerOp "Inf"    (lat_infLoos eLat) True True
supOp       = DerOp "Sup"    (lat_supLoos eLat) True True
compBinFunc :: BinFunc (FRel (Int,Int) Int Int)
compBinFunc = BinFunc "Comp" fComp
\end{code}
The controller for street one is computed as follows.
\begin{code}
deFuzz sc = UnFunc "Res" (fDown . fRRes sc)
tLightCtrA =  contrUpdEntIn  "" [(infOp,(lowT,["TR"])),
                                 (infOp,(highT,["TG"])),
                                 (infOp,(mediumT,["S"]))] $
              contrUpdEntOut "" [(infOp,(tRed,[])),
                                 (infOp,(tGreen,[])),
                                 (infOp,(stay,[]))] $
              contrUpdDefuzz (deFuzz $ fScalar eLat (80,0) unit) $
              contrUpdAggr supOp $
              contrUpdAppl compBinFunc $
              emptyContr "TLA" unit eLat
\end{code}
The construction is delivered as expected. Notice that we do not explicitly have to %%@
provide the identity function as fuzzification since \smalltt{emptyContr} uses it per %%@
definition (cf.\ Section \ref{SecModFContr}). The only things we have to set are the %%@
aggregation and defuzzification function as well as the linguistic entities for input and %%@
output, respectively. As defuzzification we choose the cut approach with the function %%@
$\Delta(y):=(\alpha^{(0.8,0)}_{I}\rres y)^{\da}$. Of course, the parameter $(0.8,0)$ %%@
strongly affects the behavior of the resulting controller and has to be investigated %%@
properly. The second component of this tuple is set to zero since we want to control %%@
street one. 

The controller for the second street is implemented analogously.
\begin{code}
tLightCtrB =  contrUpdEntIn  "" [(infOp,(lowT,["TR"])),
                                 (infOp,(highT,["TG"])),
                                 (infOp,(mediumT,["S"]))] $
              contrUpdEntOut "" [(infOp,(tRed,[])),
                                 (infOp,(tGreen,[])),
                                 (infOp,(stay,[]))] $
              contrUpdDefuzz (deFuzz $ fScalar eLat (0,80) unit) $
              contrUpdAggr supOp $
              contrUpdAppl compBinFunc $
              emptyContr "TLB" unit eLat
\end{code}
The only difference is the label and the defuzzification function. Here the first %%@
component of the scalar for the cut computation is set to zero since we want to control %%@
street two.\\
Finally, we have to put \smalltt{tLightCtrA} and \smalltt{tLightCtrB} together to one %%@
product controller. This is done by the function
\begin{code}
tLightCtr' = contrProd False True tLightCtrA tLightCtrB infOp
\end{code}
which uses the standard cross product (indicated by \smalltt{infOp}). Notice that only %%@
the range of the resulting controller has to be extended so that it steers two traffic %%@
lights. It is clear that now the critical states ($green$,$green$) and ($red$,$red$) can %%@
occur. Hence, we modify the defuzzification function of \smalltt{tLightCtr'} such that %%@
these states are excluded. We choose the most simple variant by providing a function %%@
$f:A\times A\lora A\times A$ that can be applied to the defuzzified output of %%@
\smalltt{tLightCtr'}. It shall be of the following form :
\begin{tabbing}
\qquad \= ($green$,$green$) \= $\mapsto$ \= ($red$,$green$),\\
       \> ($green$,$red$)   \> $\mapsto$ \> ($green$,$red$),\\
       \> ($red$,$green$)   \> $\mapsto$ \> ($red$,$green$),\\
       \> ($red$,$red$)     \> $\mapsto$ \> ($green$,$red$).
\end{tabbing}
Obviously, the non-critical states are not changed. But, when both streets have green, we %%@
decide to turn red on the main street and let the cars of street two pass. This is done %%@
due to fairness considerations. Otherwise it would be possible that ``L2'' never turns %%@
green (e.g., during the peak time). The traffic flow on the main street should not be %%@
affected too strong with this regulation since ``L1'' and ``L2'' are programmed to prefer %%@
street one with state $green$ if less than nine cars are waiting on street two.\\
Finally, the state ($red$,$red$) shall be turned into ($green$,$red$) which makes even %%@
clearer that we prefer the main street with state $green$.

Since $A\times A$ is represented by the list \smalltt{[20,21,20,21]} which can be thought %%@
as\\
\centerline{\smalltt{[({\em red,red}),({\em red,green}),({\em green,red}),({\em %%@
green,green})]},}
we can achieve the intended behavior by applying the matrix
\begin{small}
\[
R:=\left(\begin{array}{cccc}
           0 & 0 & 1 & 0\\
           0 & 1 & 0 & 0\\
           0 & 0 & 1 & 0\\
           0 & 1 & 0 & 0
         \end{array}\right)
\]
\end{small}
to the defuzzified output. If we denote the old defuzzification function of %%@
\smalltt{tLightCtr'} by $\Delta'$, we have the new defuzzification function %%@
$\Delta(y):=\Delta'(y);R$.

The following Haskell code realizes our intent.
\begin{code}
tLightCtr = let df' = defuzz $ header tLightCtr'
                tLT = crossProd [tLightTrg,tLightTrg]
                tEl = lat_topEl eLat
                r   = fUpd (fBot eLat tLT tLT)
                           [((0,2),tEl),((1,1),tEl),((2,2),tEl),((3,1),tEl)]
            in contrUpdDefuzz (UnFunc (unLabel df') $
                                      \x -> fComp (unFunc df' x) r) tLightCtr'
\end{code}

Now, we examplary want to have a look at the rulebase of \smalltt{tLightCtr} to see the %%@
effect of the \smalltt{contrProd} operation. We compute the rulebase as follows.
\begin{smallverb}
FContrTest> rules $ contrMelt tLightCtr
[("Inf",("LT&&LT",["TR**TR"])),("Inf",("LT&&HT",["TR**TG"])),
("Inf",("LT&&MT",["TR**S"])),("Inf",("HT&&LT",["TG**TR"])),
("Inf",("HT&&HT",["TG**TG"])),("Inf",("HT&&MT",["TG**S"])),
("Inf",("MT&&LT",["S**TR"])),("Inf",("MT&&HT",["S**TG"])),
("Inf",("MT&&MT",["S**S"]))] 
\end{smallverb}
Here we see that indeed only the output entities are extended (indicated by the %%@
concatenator \smalltt{**}). This rulebase is the direct realization of our intended %%@
controller.\\
The effect of the extended defuzzification function of \smalltt{tLightTrg} can be seen by %%@
the following computations. 
\begin{smallverb}
FContrTest>infer tLightCtr' $ fUpd (fBot eLat unit $ contrDom tLightCtr')
                              [((0,89),(100,100))]
Source: [1]
Target: [20,21,20,21]
Rel   :
[[(0,0),(0,0),(0,0),(100,100)]]

FContrTest>infer tLightCtr $ fUpd (fBot eLat unit $ contrDom tLightCtr)
                             [((0,89),(100,100))]
Source: [1]
Target: [20,21,20,21]
Rel   :
[[(0,0),(100,100),(0,0),(0,0)]]
\end{smallverb}
With the first call we compute the resulting state of the two traffic lights when we have %%@
eight cars on the main street and nine cars on street two. The result indeed indicates %%@
that this causes both ``L1'' and ``L2'' to turn green if they are controlled by %%@
\smalltt{tLightCtr'}.\\
With the second call we use the modified controller \smalltt{tLightCtr}. It computes %%@
state ($red$,$green$) for the same amount of traffic.

Now, suppose we want to introduce an additional signal which causes the traffic lights to %%@
go inactive (often indicated by a flashing yellow light). For this purpose we provide a %%@
new controller as follows. 
\begin{code}
offSrc   = unit
offCtr = let offEnt l = LingEntity l $ fTop eLat [1] offSrc
         in  contrUpdEntIn  "" [(infOp,(offEnt "OffI",["OffO"]))] $
             contrUpdEntOut "" [(infOp,(offEnt "OffO",[]))] $
             contrUpdAppl compBinFunc $
             emptyContr "Off" unit eLat
\end{code}
It does nothing more than to react on a special input signal and generate the %%@
corresponding output signal. If we combine this controller and \smalltt{tLightCtr} by
\begin{code}
extTLightCtr = contrSum True True tLightCtr offCtr supOp
\end{code}
the resulting controller can be turned off by a special signal. The following computation %%@
makes this clear.
\begin{smallverb}
FContrTest> apply extTLightCtr $ fUpd (fBot eLat unit $ contrDom extTLightCtr)
                                      [((0,100),(100,100))]
Source: [1]
Target: [20,21,20,21,1]
Rel   :
[[(0,0),(0,0),(0,0),(0,0),(100,100)]]
\end{smallverb}
The domain of \smalltt{extTLightCtr} is $(A\times A)+I$. Hence, we have $101$ entries in %%@
the domain such that the last entry represents the {\it off} signal. Using this fact, we %%@
create the crisp input fuzzy set shown above and, hence, the lights are turned off after %%@
the application of this input to the core of \smalltt{extTLightCtr}.

The computation
\begin{smallverb}
FContrTest> infer extTLightCtr $ fUpd (fBot eLat unit $ contrDom extTLightCtr)
                                      [((0,90),(100,100))]
Source: [1]
Target: [20,21,20,21,1]
Rel   :
[[(0,0),(0,0),(100,100),(0,0),(0,0)]]
\end{smallverb}
shows that the controller does not generate the signal to turn off if we have an input %%@
\linebreak representing the amount of traffic. We choose position $90$ of the domain %%@
which means that we have nine cars on street one and zero cars on street two. Hence, the %%@
controller infers the output shown above. It corresponds to the tuple ($green$,$red$), %%@
i.e., we have $green$ on street one and $red$ on street two.   
%%wholeContr = contrProd False True tLightCtrA tLightCtrB infOp 
%%timeSrc  = [(1::Int)..10] -- zhlt im 8-sekunden takt
%%timeEnt  = LingEntity "T" $ fUpd (fBot eLat [1::Int] timeSrc) [((0,i),
%%(max 0 (gran-10*i),
%%                                    max 0 (gran-10*i)))|i<-[0..9]]
%%timeCtr = contrUpdEntIn "" [(infOp,(timeEnt,["TR"]))] $
%%          contrUpdEntOut "" [(infOp,(tRed,[]))] $
%%          contrUpdAppl compBinFunc $
%%          emptyContr "TC" [1] eLat
%%tLightExtCtrA = let newDF = contrUpdDefuzz (tDeFuzz $ fScalar eLat (80,0) 
%%[1::Int])
%%                in contrProd True False (newDF tLightCtrA) (newDF timeCtr) 
%%infOp
%%tDeFuzz sc = UnFunc "TRes" (\x -> fSyQ (fConv (fDown $ fRRes sc x)) (epsi eLat 
%%tLightTrg))
%%test = fSyQ (fConv (fTop eLat [1::Int] tLightTrg)) (epsi eLat tLightTrg)
%%--preFinalContr = contrUpdEntIn
%%epsi l s = FRel l s 
%%                (replicate (2^length s) (1::Int)) $
%%                epsi' $ length s
%% where
%%    epsi' 0 = [[lat_topEl l]]
%%    epsi' 1 = [[lat_botEl l,lat_topEl l]]
%%    epsi' n = let e1 = epsi' $ n-1
%%              in zipWith (++) e1 e1 ++
%%                 [replicate (length $ head e1) (lat_botEl l)++
%%                  replicate (length $ head e1) (lat_topEl l)]
