\section{ A module for $\SL$-fuzzy relations }
\label{SecLFuzzyRels}
In this chapter we aim at a module to support the standard instantiation of Goguen %%@
categories using $\SL$-fuzzy relations. Thus, we first have to make ourselves clear what %%@
this module shall be able to do and, just as important, what it does not have to.\\ 
Of course, we need a data structure representing an $\SL$-fuzzy relation. Furthermore, %%@
the user shall be able to manipulate and construct a relation. To guarantee %%@
extendability, a very important design rule with this is to hide the implementation. %%@
Thus, we have to provide certain access routines to create an $\SL$-fuzzy relation %%@
starting from the top/bottom relation between certain source and target. Furthermore, the %%@
standard matrix operations shall be available. Within fuzzy control, new intersection and %%@
composition operators derived from lattice-ordered operator sets (loos') (cf.\ Definition %%@
\ref{DefnLoos}) play an important role. Hence, we have to provide an appropriate data %%@
structure as well as special functions for this purpose.\\
The export list
\begin{code}
module LFuzzyRel (FRel(..),Loos(..),fZip,fIncl,fComp,fJoin,fMeet,fConv,fCompLoos,
                  fOpLoos,fUp,fDown,fTop,fBot,fLRes,fRRes,fSyQ,fLResLoos,
                  fRResLoos,fSyQLoos,fDerOp,fScalar,fId,fUpd,fUpdUnBy,
                  fUpdAllUnBy,fUpdBinBy,fEntryAt,fExtSrc,fExtTrg,getFRelsBy,
                  getFRels,getAllFRelsBy,getAllFRels,lat_infLoos,lat_supLoos,
                  ) where
\end{code}
shows the provided data structures \smalltt{FRel} and \smalltt{Loos}, the basic %%@
operations (join, meet, conversion etc.), as well as their counterparts for derived %%@
operations (\smalltt{fOpLoos} and  \smalltt{fCompLoos} etc.), the top and bottom %%@
relation, some functions for manipulating relations componentwise (\smalltt{fUpd}, %%@
\smalltt{fUpdUnBy}, etc.) and different routines to compute all $\SL$-fuzzy relations %%@
fulfilling a certain predicate between given source and target ({\smalltt{getFRelsBy}, %%@
\smalltt{getAllFRelsBy}, etc.). The last functions become interesting when we instantiate %%@
relational categories later on.

This basic functionality suffices for our purposes and leaves space for future %%@
extensions. These extensions could, for example, go into deep with $\SL$-fuzzy relations %%@
and provide an implementation of special operations like fuzzy implication or fuzzy %%@
negation.

Now, we are ready to switch to the implementation. Since the entries of our $\SL$-fuzzy %%@
relations have to constitute a complete Brouwerian lattice, we obviously have to import %%@
the \smalltt{Lattice} module. Furthermore, some special list operations are needed later %%@
on. 
\begin{code}
import Lattice
import List (transpose,nub)
\end{code}
The next step is the \smalltt{FRel} data structure. Several design decisions have to be %%@
made here.
\begin{code}
data FRel e obj1 obj2 =  FRel { lat   :: Lat e
                               ,src   :: [obj1]
                               ,trg   :: [obj2]
                               ,rel   :: [[e]] }                            
\end{code}
An $\SL$-fuzzy relation consists of its underlying entry lattice \smalltt{lat}, its %%@
source and target (\smalltt{src}, \smalltt{trg}) and the real relation \smalltt{rel} %%@
represented as a matrix.\\
Notice that we deliberately do not use the \smalltt{Set} module to describe source and %%@
target, respectively. The reason is that the interpretability of our relations would %%@
suffer because we would not know how the rows and columns are labeled. Thus, %%@
\smalltt{src} resp.\ \smalltt{trg} can be seen as the labeling of the given relation.

As mentioned above, we use the matrix representation for our $\SL$-fuzzy relations and %%@
implement it by a list of lists of entries. Another possibility would have been to use %%@
the preimplemented \smalltt{Array} module, which has the advantage that we can rely on %%@
predefined functions and have a comprehensive error detection. But, there are several %%@
drawbacks of this approach. The first one is that the range of an array has to be %%@
indexable, i.e., any type we would use for \smalltt{src} and \smalltt{trg} had to %%@
implement the \smalltt{Ix} type class. For \smalltt{Double} and \smalltt{Rational} values %%@
this is at least difficult. The second (even worse) drawback is that we could only use %%@
coherent source and target ranges, respectively. Thus, the source $\{1,2,3,5\}$ would not %%@
be allowed. At least with the automated construction of relations this causes problems. %%@
Consider, for example, two relations $R:A\lora B$ and $S:A\lora C$ with $B:=\{1,2,3\}$ %%@
and $C:=\{7,8,9,10\}$. Obviously, we have to represent them by   
\begin{smallverb}
relR = { lat = ...     relS = { lat = ...      
        ,src = ...             ,src = ...
        ,trg = (1,3)           ,trg = (7,10)
        ,rel = ... }           ,rel = ... }
\end{smallverb}
if we use the array implementation. But, if we want to construct the direct sum, we get %%@
problems computing the \smalltt{trg} parameter of the resulting injections. Obviously, %%@
\smalltt{(1,10)} is not allowed. Since the construction of \smalltt{trg} would only be %%@
possible with considerable efforts and the interpretability of the resulting relations %%@
would suffer anyway, we use the list representation. Now, one could still use the %%@
\smalltt{Array} module to represent the matrix by mapping source and target to suitable %%@
array ranges (e.g., each element could be mapped to its list index). But, tests showed %%@
that the \smalltt{Array} module is slower than the list operations. All this drove us to %%@
define \smalltt{FRel} as shown above. Of course, several consistency tests for matrices %%@
have to be provided with our approach. 

Later on, we need \smalltt{Show} and \smalltt{Eq} instances of \smalltt{FRel}.
\begin{code}
instance (Show e,Show obj1, Show obj2) => Show (FRel e obj1 obj2) where
  show r = "Source: "++ (show $ src r) ++ "\n" ++
           "Target: "++ (show $ trg r) ++ "\n" ++
           "Rel   : "++ "\n" ++ (show $ rel r)++"\n"
		   
instance (Eq e, Eq obj1, Eq obj2) => Eq (FRel e obj1 obj2) where
  (==) rel1 rel2 = src rel1 == src rel2 &&
                   trg rel1 == trg rel2 &&
                   rel rel1 == rel rel2
\end{code}
The \smalltt{Eq} instance again makes clear that two relations are only equal if they %%@
have equally labeled rows and columns as well as equal matrix entries.

The data structure for loos' is implemented as follows.
\begin{code}
data Loos el = Loos { loos_lat :: Lat el
                     ,loos_op  :: el -> el -> el
                     ,loos_e   :: el
                     ,loos_z   :: el }
\end{code}
It is a direct realization of the underlying definition whereas \smalltt{e} is the %%@
neutral and \smalltt{z} the zero element. This data structure induces some standard %%@
instantiations since conventional join and meet, respectively, constitute a loos.
\begin{code}
lat_infLoos,lat_supLoos :: Lat el -> Loos el
lat_infLoos l = Loos { loos_lat = l
                      ,loos_op  = lat_inf l
                      ,loos_e   = lat_topEl l
                      ,loos_z   = lat_botEl l }
                      
lat_supLoos l = Loos { loos_lat = l
                      ,loos_op  = lat_sup l
                      ,loos_e   = lat_botEl l
                      ,loos_z   = lat_topEl l }
\end{code}
Obviously, the neutral and zero elements, respectively, are dual with %%@
\smalltt{lat\_infLoos} and \smalltt{lat\_supLoos}.
   
In the following some auxiliary functions are defined. Notice, that they are not %%@
exported.\\
Later on, we want to provide different functions to access certain entries of $\SL$-fuzzy %%@
relations and to set/update/read them. Thus, we provide the functions
\begin{code}
updateAtBy :: (a -> a) -> Int -> Int -> [[a]] -> [[a]]
updateAtBy f i j mat = let
                         matSplitted = splitAt i mat
                         rowSplitted = splitAt j (head $ snd matSplitted)                     
                       in (fst matSplitted)++[(fst rowSplitted)++
                          (f (head $ snd rowSplitted) : tail(snd rowSplitted))]++
                          (tail $ snd matSplitted)

insertAt :: Int -> Int -> a -> [[a]] -> [[a]]
insertAt i j el = updateAtBy (const el) i j
\end{code}
for this purpose. They are different in the manner that \smalltt{insertAt} only inserts %%@
an element \smalltt{el} in a matrix \smalltt{mat} at position (\smalltt{i},\smalltt{j}}), %%@
and \smalltt{updateAtBy} updates the element in question using the unary function %%@
\smalltt{f}.

For error handling we provide the following error message to inform the user properly.
\begin{code}
srcTrgError :: String
srcTrgError = "Underlying sources/targets not suitable !"
\end{code}
Hence, we are ready to implement the standard operations for $\SL$-fuzzy relations. To %%@
avoid name space conflicts, all standard functions for $\SL$-fuzzy relations will be %%@
started by \smalltt{f}.\\
The inclusion operation is delivered by the following function.
\begin{code}
fIncl :: (Eq obj1, Eq obj2) => FRel e obj1 obj2 -> FRel e obj1 obj2 -> Bool 
fIncl rel1 rel2 = let l    = lat rel1
                  in if (src rel1)==(src rel2) && (trg rel1) == (trg rel2) then
                       fIncl' l (rel rel1) (rel rel2)
                     else error srcTrgError
                  where
                   fIncl' l []       []               = True
                   fIncl' l ([]:r1s) ([]:r2s)         = fIncl' l r1s r2s
                   fIncl' l ((x:xs):r1s) ((y:ys):r2s) = (lat_lEq l x y) && 
                                                       fIncl' l (xs:r1s) (ys:r2s)
\end{code}
The inclusion check is done componentwise and immediately finished when two elements are %%@
found that are not in the corresponding order relation. Notice that our error detection %%@
is restricted to determining whether the matrices are compatible with respect to their %%@
sources and targets. We do not test whether the matrix given in \smalltt{rel} really %%@
corresponds to the given source and target, respectively. This has to be done manually by %%@
the user after constructing the $\SL$-fuzzy relation. To do so, the test functions %%@
introduced later on can be used. With this restriction we are able to guarantee %%@
consistency and acceptable run times.
 
The next step is to provide join, meet and composition operations. We start with %%@
operations derived from loos'. Since standard join and meet constitute a loos we then can %%@
rely on these functions.  
\begin{code}
fZip :: (Eq obj1, Eq obj2) => 
        (e -> e -> e) -> FRel e obj1 obj2 -> FRel e obj1 obj2 -> FRel e obj1 obj2
fZip f rel1 rel2 = let l = lat rel1 
                   in 
                     if (src rel1)==(src rel2) && (trg rel1) == (trg rel2) then
                       FRel l (src rel1) (trg rel2) $ 
                            zipWith (zipWith f) (rel rel1) (rel rel2)
                     else error srcTrgError

fOpLoos :: (Eq obj1, Eq obj2) =>
           Loos e -> FRel e obj1 obj2 -> FRel e obj1 obj2 -> FRel e obj1 obj2
fOpLoos loos = fZip (loos_op loos)

fCompLoos :: (Eq obj2) =>
           Loos e -> FRel e obj1 obj2 -> FRel e obj2 obj3 -> FRel e obj1 obj3
fCompLoos loos rel1 rel2 =
  let l = lat rel1 
  in if (trg rel1)==(src rel2)then
       FRel l (src rel1) (trg rel2)
              [ [foldr1 (lat_sup l) $ zipWith (loos_op loos) x [y!!i|y<-rel rel2]
                 | i<-[0..length (head $ rel rel2)-1] ] 
                | x<-rel rel1]
     else error srcTrgError
\end{code}
The key functionality for derived join/meet operations is established in \smalltt{fZip} %%@
which takes two $\SL$-fuzzy relations and then concatenates them using a given binary %%@
function \smalltt{f}. Hence,  \smalltt{fOpLoos} can use \smalltt{fZip} parametrized by %%@
the loos operator \smalltt{loos\_op} of \smalltt{loos}. Finally, the derived composition %%@
operator is delivered by \smalltt{fCompLoos}.

Using \smalltt{lat\_infLoos} and \smalltt{lat\_supLoos} from above, we are now able to %%@
provide the standard operations.
\begin{code}
fJoin,fMeet :: (Eq obj1, Eq obj2) => 
                FRel e obj1 obj2 -> FRel e obj1 obj2 -> FRel e obj1 obj2
fJoin rel1 = fOpLoos (lat_supLoos $ lat rel1) rel1

fMeet rel1 = fOpLoos (lat_infLoos $ lat rel1) rel1

fComp :: (Eq obj2) => FRel e obj1 obj2 -> FRel e obj2 obj3 -> FRel e obj1 obj3 
fComp rel1 = fCompLoos (lat_infLoos $ lat rel1) rel1 
\end{code}
Conversion and the up/down operators are implemented as follows.
\begin{code}
fConv :: FRel e obj1 obj2 -> FRel e obj2 obj1
fConv rel1 = FRel (lat rel1) (trg rel1) (src rel1) (transpose $ rel rel1)

fUp,fDown :: FRel e obj1 obj2 -> FRel e obj1 obj2
fUp rel1 = let l     = lat rel1
           in FRel l (src rel1) (trg rel1) $
                     map (map (\x -> if   lat_lEq l x $ lat_botEl l then x
                                     else lat_topEl l ) ) $ rel rel1

fDown rel1 = let l     = lat rel1
             in FRel l (src rel1) (trg rel1) $
                       map (map (\x -> if   lat_lEq l (lat_topEl l) x then x
                                       else lat_botEl l ) ) $ rel rel1
\end{code}
Now, we want to introduce some standard relations which have to be used to construct any %%@
other relation componentwise.
\begin{code}
fScalar :: Lat e -> e -> [obj1] -> FRel e obj1 obj1
fScalar e el s = let lS  = length s
                     botEl = lat_botEl e
                 in FRel e s s [replicate i botEl ++ 
                                           (el : replicate (length s-i-1) botEl) 
                                | i<-[0..length s-1]]

fId :: Lat e -> [obj1] -> FRel e obj1 obj1
fId e = fScalar e (lat_topEl e)

fTop,fBot :: Lat e -> [obj1] -> [obj2] -> FRel e obj1 obj2
fTop e s t = FRel e s t $ replicate (length s) 
                                    (replicate (length t) $ lat_topEl e)

fBot e s t = FRel e s t $ replicate (length s) 
                                    (replicate (length t) $ lat_botEl e)
\end{code}
From the function names it is obvious which kind of relation they implement. All routines %%@
have to get the underlying lattice as well as the source and target object. Since source %%@
and target are identical with the scalars and the identity relation, we need only one %%@
parameter. Notice that the set of all scalars for a given source plays an important role %%@
within Goguen categories since it is isomorphic to the elements of the entry lattice.

As indicated by the $\alpha$-cut Theorem (cf.\ Theorem \ref{ThmACut}), the left and right %%@
residual as well as the symmetric quotient of two given $\SL$-fuzzy relations are %%@
important operations within Goguen categories. Before we can implement them, we have to %%@
undergo some efficiency considerations. The following lemma characterizes the entries of %%@
the left residual. Notice that the underlying definition $Q;R\sqsse S\lra Q\sqsse (S/R)$ %%@
indicates that new residued operations can be derived from lattice-ordered semigroups. %%@
Hence, we consider the general case.
\begin{lem}
\label{LemRes}
Let $R:B\lora C$ and $S:A\lora C$ be two $\SL$-fuzzy relations over $\SL$. Furthermore, %%@
let $*$ be a loos based derived operation on $\SL$. Then the left residual $S/_{*}R$ can %%@
be computed componentwise by $(S/_{*}R)(x,y)=\bbigvee\{a\in\SL\ |\ \forall z: %%@
a*R(y,z)\leq S(x,z)\}$.
\end{lem}
\begin{proof}[{\bf Proof.}]
We show \\
\centerline{$\forall x,z: (Q;_{*}R)(x,z)\leq S(x,z)\lra \forall x,y: %%@
Q(x,y)\leq\bbigvee\{a\in\SL\ |\ \forall z: a*R(y,z)\leq S(x,z)\}$ }. The assertion %%@
immediately follows from
\begin{tabbing}
\qquad \= \qquad \= $\forall x,y: Q(x,y)\leq\bbigvee\{a\in\SL\ |\ \forall z: a* %%@
R(y,z)\leq S(x,z)\}$ \\
       \> $\lra$ \> $\forall x,y,z: Q(x,y)* R(y,z)\leq S(x,z)$\\
       \> $\lra$ \> $\forall x,z: \bbigvee_{y}(Q(x,y)* R(y,z))\leq S(x,z)$\\
       \> $\lra$ \> $\forall x,z: (Q;_{*}R)(x,z)\leq S(x,z)$.
\end{tabbing}
\end{proof}
Since $Q\rres_{*} S=(S^{\conv}/_{*}Q^{\conv})^{\conv}$ we automatically have that %%@
$(Q\rres_{*} S)(y,z)$ can be computed by $\bbigvee\{a\in\SL\ |\ \forall x: a*Q(x,y)\leq %%@
S(x,z)\}$.

With the implementation, we again first introduce functions for derived operations and %%@
then give the standard variants as a special case.
\begin{code}
fLResLoos :: (Eq obj1, Eq obj2, Eq obj3) => Loos e ->
             FRel e obj1 obj3 -> FRel e obj2 obj3 -> FRel e obj1 obj2
fLResLoos loos t s = 
   let targ = src s
       sour = src t
       l    = lat t
       sup  = lat_sup l
       lEq  = lat_lEq l
       f    = loos_op loos
   in FRel { lat = l
            ,src = sour
            ,trg = targ
            ,rel = part (length targ) 
                        [let sl  = (rel s !! y)
                             tl  = (rel t !! x)
                             els = [ el | el<-lat_elements l, and $ 
                                          zipWith (\a b -> lEq (f el a) b) sl tl]
                         in if length els>1 then foldl1 sup els
                            else head els
                          | x<-[0..length sour-1], y<-[0..length targ-1]] }
   where
     part i [] = []
     part i xs = let (s1,s2) = splitAt i xs 
                 in s1 : part i s2

fRResLoos,fSyQLoos :: (Eq obj1, Eq obj2, Eq obj3) =>  Loos e ->
             FRel e obj1 obj2 -> FRel e obj1 obj3 -> FRel e obj2 obj3
fRResLoos loos s t = fConv $ fLResLoos loos (fConv t) (fConv s)

fSyQLoos loos s t = fMeet (fRResLoos loos s t) $ fLResLoos loos (fConv s) (fConv t)

fLRes :: (Eq obj1, Eq obj2, Eq obj3) => 
         FRel e obj1 obj3 -> FRel e obj2 obj3 -> FRel e obj1 obj2
fLRes t = fLResLoos (lat_infLoos $ lat t) t

fRRes,fSyQ :: (Eq obj1, Eq obj2, Eq obj3) =>  
         FRel e obj1 obj2 -> FRel e obj1 obj3 -> FRel e obj2 obj3
fRRes s = fRResLoos (lat_infLoos $ lat s) s

fSyQ s t = fMeet (fRRes s t) $ fLRes (fConv s) (fConv t)
\end{code}
Obviously, the realization of \smalltt{fLResLoos} directly corresponds to Lemma %%@
\ref{LemRes}. The other functions can rely on it so that no separate implementations are %%@
necessary. Notice that with \smalltt{fSyQLoos} and \smalltt{fSyq} we definitely have to %%@
use the \smalltt{fMeet} function to concatenate the left resp.\ right residual. This %%@
comes due to the fact that the {\em and} operator of the underlying definition  $X\sqsse %%@
syQ(Q,S)$, iff $X\sqsse (Q\rres S)$ {\em and} $X\sqsse (Q^{\conv}/S^{\conv})$ has to be %%@
represented by \smalltt{fMeet}.\\
As we saw in the \smalltt{Goguen} module, we need a standard routine that generates %%@
certain derived operations within Goguen categories. This is done by the following %%@
function.
\begin{code}
fDerOp :: (Eq obj) => String -> Loos e -> 
                      FRel e obj obj -> FRel e obj obj -> FRel e obj obj
fDerOp l =
  case l of
     "Meet" -> fOpLoos
     "Comp" -> fCompLoos
     "LRes" -> fLResLoos 
     "RRes" -> fRResLoos
     "syQ"  -> fSyQLoos
\end{code}
The label \smalltt{l} indicates which kind of derived operation shall be generated. %%@
According to the definition of \smalltt{gog\_derOp}, the user only has to provide the %%@
loos to have a derived operation within Goguen categories.

Up to now, we are only able to handle given $\SL$-fuzzy relations. But, we need some %%@
suitable functions to construct them in a componentwise manner. The most simple operation %%@
is to read an element at a given position in the relation.
\begin{code}
fEntryAt :: FRel e obj1 obj2 -> (Int,Int) -> e
fEntryAt rel1 (i,j) = ((rel rel1) !! i) !! j
\end{code}
The user has to give the row \smalltt{i} and the column \smalltt{j} of the entry he wants %%@
to read. Obviously, \smalltt{i} and \smalltt{j} have to describe the position in the %%@
matrix of the relation \smalltt{rel1} so that we can rely on the \smalltt{!!} operator. %%@
Notice that the row and column count, respectively, start with $0$.

The following functions can be used to update given $\SL$-fuzzy relations. This can be %%@
done threefold. First, the user can give a list of tuples, each tuple containing an index %%@
pair and the element that shall be set. This is covered by \smalltt{fUpd} which relies on %%@
the above defined function \smalltt{insertAt}.
\begin{code}
fUpd :: FRel e obj1 obj2 -> [((Int,Int),e)] -> FRel e obj1 obj2 
fUpd rel1 els = FRel (lat rel1) (src rel1) (trg rel1) $ insertAll els (rel rel1)
                where
                  insertAll []             m = m
                  insertAll (((i,j),x):xs) m = insertAt i j x $ insertAll xs m
\end{code}
The second variant is to update an element in the matrix by a unary function \smalltt{f}. %%@
Here we can differentiate two cases. If one explicitly wants to give the indices that %%@
shall be updated, one can use \smalltt{fUpdUnBy}. If the whole matrix shall be updated, %%@
the function \smalltt{fUpdAllUnBy}, which makes use of the above defined function %%@
\smalltt{updateAtBy}, is to use.
\begin{code}
fUpdUnBy :: (e -> e) -> FRel e obj1 obj2 -> [(Int,Int)] -> FRel e obj1 obj2
fUpdUnBy f rel1 ixs = FRel (lat rel1) (src rel1) (trg rel1) $ 
                           updateAllBy f ixs (rel rel1)
                      where
                        updateAllBy _ []          m = m
                        updateAllBy f ((i,j):ixs) m = updateAtBy f i j $ 
					                                  updateAllBy f ixs m

fUpdAllUnBy :: (e -> e) -> FRel e obj1 obj2 -> FRel e obj1 obj2
fUpdAllUnBy f rel1 = FRel (lat rel1) (src rel1) (trg rel1)
                          [[f y | y<-x] | x<-rel rel1]
\end{code}
Notice that these three functions would suffice to construct and manipulate $\SL$-fuzzy %%@
relations quite comfortably. But, one may want to update certain entries by a binary %%@
function applied to the old entry and another (given) element. A typical scenario is the %%@
``partial join'' operation of two matrices where only some selected fields shall be %%@
joined. We support this kind of operation by the function   
\begin{code}
fUpdBinBy :: (e -> e -> e) -> FRel e obj1 obj2 -> [((Int,Int),e)] -> 
             FRel e obj1 obj2
fUpdBinBy f rel1 els = FRel (lat rel1) (src rel1) (trg rel1) $ 
                            updateAllBy f els (rel rel1)
                       where
                          updateAllBy _ []             m = m
                          updateAllBy f (((i,j),e):xs) m = updateAtBy (f e) i j $ 
                                                           updateAllBy f xs m
\end{code}
which again makes use of \smalltt{updateAtBy}. It takes a binary function together with a %%@
list of tuples, each tuple containing an index pair and an element, and updates the %%@
entries in question. Obviously, \smalltt{fUpdBinBy} could be easily reduced to %%@
\smalltt{fUpdUnBy}. But, for efficiency considerations, we again implement %%@
\smalltt{updateAllBy} with the only difference that \smalltt{f} here is a binary (instead %%@
of unary) function.

Thus, we have the basic functionality. Later on, we want to instantiate different kinds %%@
of relational categories using this module. From the underlying definitions we know that %%@
the relations (morphisms) have to form certain kinds of structures. Hence, we can support %%@
these instantiations by providing functions that automatically compute certain %%@
$\SL$-fuzzy relations between given source and target. The function
\begin{code}
getFRelsBy :: (FRel e obj1 obj2 -> Bool) -> Lat e -> [obj1] -> [obj2] -> 
              [FRel e obj1 obj2]
getFRelsBy pr e s t = 
    let lS = length s
        lT = length t-1
    in filter pr [FRel e s t x | x <- map (split $ lT+1) 
	                                      (allocate (length s*length t-1) e)]
    where
      allocate 0 e = map (:[]) $ lat_elements e
      allocate n e = [ o:m | o<-lat_elements e, m<-allocate (n-1) e]
      split _ [] = []
      split i xs = let s = splitAt i xs
                   in fst s : split i (snd s)
\end{code}
takes a unary predicate \smalltt{pr} on the set of all $\SL$-fuzzy relations between %%@
source \smalltt{s} and target \smalltt{t} and an entry lattice \smalltt{e}. It then %%@
returns a list of all $\SL$-fuzzy relations between \smalltt{s} and \smalltt{t} %%@
fulfilling \smalltt{pr}. This is done with the help of two functions. First, %%@
\smalltt{allocate} constructs all possibilities to combine \smalltt{length s * length t} %%@
elements of the entry lattice. It returns a list for every combination. Obviously, the %%@
length of such a list exactly corresponds to the number of entries in a relation between %%@
\smalltt{s} and \smalltt{t}. To achieve our matrix representation, we then use the %%@
\smalltt{split} function to split the lists. The result then can be used to construct the %%@
final $\SL$-fuzzy relations. At the end we filter out the relations fulfilling %%@
\smalltt{pr}.\\  
The standard variant of \smalltt{getFRelsBy} returns all $\SL$-fuzzy relations.
\begin{code}                   
getFRels :: Lat e -> [obj1] -> [obj2] -> [FRel e obj1 obj2]
getFRels = getFRelsBy $ const True
\end{code}
Good examples for a special predicates are ``is crisp'' or ``is total''.

The function 
\begin{code}
getAllFRelsBy :: Eq obj => 
               (FRel e obj obj -> Bool) -> Lat e -> [[obj]] -> [[FRel e obj obj]]
getAllFRelsBy pr entry objs = let os = nub objs 
                              in [getFRelsBy pr entry src trg | src<-os,trg<-os]
\end{code}
goes a step further. It takes a list of sources and targets (\smalltt{objs}) and computes %%@
all $\SL$-fuzzy relations fulfilling the predicate \smalltt{pr} between any two elements %%@
out of \smalltt{objs}. With this it relies on \smalltt{getFRelsBy}. Later on, %%@
\smalltt{getAllFRelsBy} can be used in quite a comfortable way to instantiate certain %%@
relational categories. Furthermore, we see a little demerit of Haskell's strong typing. %%@
We can only use one type of sources/targets because we put them all together into one %%@
list. The same is true for the resulting $\SL$-fuzzy relations.\\
Again, we provide a standard version that computes all relations.
\begin{code}
getAllFRels :: Eq obj => Lat e -> [[obj]] -> [[FRel e obj obj]]
getAllFRels = getAllFRelsBy $ const True
\end{code}
Before we come to the end, we want to provide two auxiliary functions which will be used %%@
within our module for fuzzy controllers, later on.
\begin{code}
fExtSrc,fExtTrg :: [[obj]] -> FRel e obj obj -> [[obj]] -> 
                                   FRel e obj obj
fExtSrc usrcs r lsrcs = FRel (lat r) (concat usrcs++src r++concat lsrcs) (trg r) 
                             (rel (fBot (lat r) (concat usrcs) $ trg r)++rel r++
                              rel (fBot (lat r) (concat lsrcs) $ trg r))
fExtTrg ltrgs r rtrgs = 
   FRel (lat r) (src r) (concat ltrgs++trg r++concat rtrgs) 
        (zipWith (++)         (rel $ fBot (lat r) (src r) $ concat ltrgs) $
         zipWith (++) (rel r) (rel $ fBot (lat r) (src r) $ concat rtrgs))
\end{code}
Their manner of function shall be explained with an example. Suppose, $R:A_{1}\lora B$ is %%@
a relation and $A_{2}$ is a set. If we want to extend $R$ to be of type $A_{1}+A_{2}\lora %%@
B$, we have to use the induced crisp injection $\iota:A_{1}\lora A_{1}+A_{2}$ and compute %%@
$\iota_{1}^{\conv};R$. But, this computation does nothing more than to add $|A_{2}|$ zero %%@
rows to $R$. Hence, we can avoid to use the composition operator and, thus, can compute %%@
the term rather efficiently. This is exactly what is done by \smalltt{fExtSrc}. The %%@
parameter \smalltt{usrcs} is a list of objects and causes \smalltt{fExtSrc} to extend the %%@
given relation \smalltt{r} to the upper side. Analogously, \smalltt{lsrcs} is used to %%@
extend \smalltt{r} to the lower side. The function \smalltt{fExtTrg} does exactly the %%@
same, but extends the target of \smalltt{r}. This corresponds to adding zero columns to %%@
the left resp.\ right side of \smalltt{r}.

Last but not least, we provide a consistency check.
\begin{code}
check :: FRel e obj1 obj2 -> Bool
check r = length (rel r) == length (src r) &&
          and (map ((== length (trg r)).length) $ rel r)  
\end{code}
We test whether the given matrix is consistent with the given source and target of the %%@
$\SL$-fuzzy relation, respectively.

At this point the implemented functions are sufficient for our purposes. We are now ready %%@
to provide suitable lattice instantiations as well as instantiations of the relational %%@
categories.%%@
