\section{ Lattices of $\SL$-fuzzy relations }
In this section we deliver predefined functions to instantiate certain kinds of lattices %%@
of $\SL$-fuzzy relations.
\begin{code}
module LFuzzyRelLattices (fRelPoSetBy   ,fRelPoSet   ,fRelPoSetList   ,
                          fRelUSemiLatBy,fRelUSemiLat,fRelUSemiLatList,
                          fRelLSemiLatBy,fRelLSemiLat,fRelLSemiLatList,
                          fRelLatBy     ,fRelLat     ,fRelLatList      ) where

import LFuzzyRel
import Lattice
\end{code}
The export list shows the structures that can be generated. To do so, we have to import %%@
the modules shown above.\\
We start with the weakest structure \smalltt{PoSet}. 
\begin{code}
fRelPoSetBy :: (Eq obj1, Eq obj2) => 
               (FRel e obj1 obj2 -> Bool) -> Lat e -> 
			   [obj1] -> [obj2]  -> PoSet (FRel e obj1 obj2)
fRelPoSetBy pr e s t = PoSet { poSet_isElem  = \x -> src x==s && trg x==t && pr x
                              ,poSet_elements = getFRelsBy pr e s t
                              ,poSet_lEq      = fIncl }

fRelPoSet :: (Eq obj1, Eq obj2) => 
             Lat e -> [obj1] -> [obj2] -> PoSet (FRel e obj1 obj2)
fRelPoSet = fRelPoSetBy $ const True

fRelPoSetList :: (Eq e, Eq obj1, Eq obj2) => 
                 Lat e -> [FRel e obj1 obj2] -> PoSet (FRel e obj1 obj2)
fRelPoSetList e rs = PoSet { poSet_isElem   = flip elem $ rs
                            ,poSet_elements = rs
                            ,poSet_lEq      = fIncl }           
\end{code}
As shown here, posets can be generated threefold. To get the poset of all relations %%@
between source \smalltt{s} and target \smalltt{t} fulfilling a unary predicate %%@
\smalltt{pr}, one can use the function \smalltt{fRelPoSetBy}. It relies on %%@
\smalltt{getFRelsBy} to generate the requested $\SL$-fuzzy relations. The test whether a %%@
certain relation belongs to this poset is done by checking source and target of the %%@
relation in question as well as the predicate property. As another possibility, one could %%@
check whether the relation is in the element list \smalltt{poSet\_elements}. But, we %%@
think testing a single predicate is faster in the most general cases (e.g., for the %%@
crispness property). Furthermore, we avoid a more detailed parametrization since a user %%@
can easily instantiate ``special'' posets himself.\\
The second variant then is the standard version. It delivers the poset of all $\SL$-fuzzy %%@
relations between given source and target.\\
The last function \smalltt{fRelPoSetList} takes a list of the relations that shall belong %%@
to the created poset. It then delivers the requested structure. Notice that we do not %%@
perform any consistency checks with the given relation list.

Having this, we can proceed with lattice structures. The construction of upper %%@
semilattices is shown with the next three functions.
\begin{code}
fRelUSemiLatBy :: (Eq obj1, Eq obj2) => 
                  (FRel e obj1 obj2 -> Bool) -> Lat e -> 
				  [obj1] -> [obj2]  -> USemiLat (FRel e obj1 obj2)
fRelUSemiLatBy pr e s t = USemiLat { uSemiLat_poSet = fRelPoSetBy pr e s t
                                    ,uSemiLat_sup   = fJoin
                                    ,uSemiLat_topEl = fTop e s t }

fRelUSemiLat :: (Eq obj1, Eq obj2) =>
                Lat e -> [obj1] -> [obj2] -> USemiLat (FRel e obj1 obj2)
fRelUSemiLat = fRelUSemiLatBy $ const True                                    

fRelUSemiLatList :: (Eq e, Eq obj1, Eq obj2) => 
                    Lat e -> [FRel e obj1 obj2] -> USemiLat (FRel e obj1 obj2)
fRelUSemiLatList e rs = let
                          s = if null rs then [] else src $ head rs
                          t = if null rs then [] else trg $ head rs
                        in 
                          USemiLat { uSemiLat_poSet = fRelPoSetList e rs
                                    ,uSemiLat_sup   = fJoin
                                    ,uSemiLat_topEl = fTop e s t }
\end{code}
Notice that with \smalltt{fRelUSemiLatList} we have to catch the case that the relation %%@
list is empty. In this situation we take the empty list as source and target, %%@
respectively, to construct the top element \smalltt{uSemiLat\_topEl}.

From the definitions of the underlying data types we know that we need to compute the %%@
atoms for lower semi lattices and lattices as well as the join resp.\ meet-irreducible %%@
elements for lattices. To do so, we first have to make ourselves clear which $\SL$-fuzzy %%@
relations over a certain lattice are irreducible and atoms, respectively. Obviously, an %%@
$\SL$-fuzzy relation is an atom or irreducible, respectively, if and only if all %%@
occurrent entries are. \\
We support this by providing the auxiliary function \smalltt{specialEls}.    
\begin{code}
specialEls :: Lat e -> FRel e obj1 obj2 -> [e] -> [FRel e obj1 obj2] 
specialEls e bt xs = [fUpd bt [((i,j),x)] 
				     | i<-[0..length (src bt)-1],j<-[0..length (trg bt)-1],x<-xs]
\end{code}
First it takes the entry lattice \smalltt{e}. For efficiency reasons the parameter %%@
\smalltt{bt} is useful. If we consider the Boolean lattice with atoms $a$ and $b$ with %%@
least/greatest element $0$/$1$, the lattice of all $\SL$-fuzzy relations between source %%@
$\{1,2\}$ and target $\{3,4\}$ has the following atoms and join-irreducible elements
\begin{small}
\[ \left(\begin{array}{cc}
                a & 0\\
                0 & 0
         \end{array}\right),\ \left(\begin{array}{cc}
                       0 & a\\
                       0 & 0
                \end{array}\right),\ \left(\begin{array}{cc}
                       0 & 0\\
                       a & 0
                \end{array}\right),\ \left(\begin{array}{cc}
                       0 & 0\\
                       0 & a
                \end{array}\right),\ \\
\left(\begin{array}{cc}
                b & 0\\
                0 & 0
         \end{array}\right),\ \left(\begin{array}{cc}
                       0 & b\\
                       0 & 0
                \end{array}\right),\ \left(\begin{array}{cc}
                       0 & 0\\
                       b & 0
                \end{array}\right),\ \left(\begin{array}{cc}
                       0 & 0\\
                       0 & b
                \end{array}\right). \]
\end{small}
The meet-irreducible elements are given by
\begin{small} 
\[ \left(\begin{array}{cc}
                a & 1\\
                1 & 1
         \end{array}\right),\ \left(\begin{array}{cc}
                       1 & a\\
                       1 & 1
                \end{array}\right),\ \left(\begin{array}{cc}
                       1 & 1\\
                       a & 1
                \end{array}\right),\ \left(\begin{array}{cc}
                       1 & 1\\
                       1 & a
                \end{array}\right),\ \\
\left(\begin{array}{cc}
                b & 1\\
                1 & 1
         \end{array}\right),\ \left(\begin{array}{cc}
                       1 & b\\
                       1 & 1
                \end{array}\right),\ \left(\begin{array}{cc}
                       1 & 1\\
                       b & 1
                \end{array}\right),\ \left(\begin{array}{cc}
                       1 & 1\\
                       1 & b
                \end{array}\right). \]
\end{small}
Hence, for the atoms/join-irreducible elements it is useful to successively update the %%@
bottom relation between given source and target. For the meet-irreducible elements it is %%@
better to use the full relation. Which one to use is told by the parameter \smalltt{bt}. %%@
Last but not least, the atoms/join- resp.\ meet-irreducible elements of the entry lattice %%@
\smalltt{e} have to be given in the parameter \smalltt{xs} so that the computation can be %%@
done.

The next functions generate lower semilattices in analogy to the poset functions above.
\begin{code}
fRelLSemiLatBy :: (Eq obj1, Eq obj2) => 
                  (FRel e obj1 obj2 -> Bool) -> Lat e -> 
                  [obj1] -> [obj2] -> LSemiLat (FRel e obj1 obj2)
fRelLSemiLatBy pr e s t = 
       LSemiLat { lSemiLat_poSet = fRelPoSetBy pr e s t
                 ,lSemiLat_inf   = fMeet
                 ,lSemiLat_botEl = fBot e s t
                 ,lSemiLat_atomS = filter pr (specialEls e (fBot e s t) $ 
                                                         lat_atomS e) }

fRelLSemiLat :: (Eq obj1, Eq obj2) =>
                Lat e -> [obj1] -> [obj2] -> LSemiLat (FRel e obj1 obj2)
fRelLSemiLat = fRelLSemiLatBy $ const True

fRelLSemiLatList :: (Eq e, Eq obj1, Eq obj2) => 
                    Lat e -> [FRel e obj1 obj2] -> LSemiLat (FRel e obj1 obj2)
fRelLSemiLatList e rs = 
      let
        s = if null rs then [] else src $ head rs
        t = if null rs then [] else trg $ head rs
      in 
        LSemiLat { lSemiLat_poSet = fRelPoSetList e rs
                  ,lSemiLat_inf   = fMeet
                  ,lSemiLat_botEl = fBot e s t
                  ,lSemiLat_atomS = specialEls e (fBot e s t) $ lat_atomS e }
\end{code}
There is nothing special to mention except the fact that we have to take care with the %%@
computation of the atoms with \smalltt{fRelLSemiLatBy}. Since the $\SL$-fuzzy relations %%@
between given source and target fulfilling a certain predicate may form a true %%@
sub(semi)lattice of the lattice of all $\SL$-fuzzy relations, we have to filter out the %%@
atoms fulfilling the predicate \smalltt{pr}.

Analogously we provide standard routines to determine the lattice of certain $\SL$-fuzzy %%@
relations between given source and target. 
\begin{code}
fRelLatBy :: (Eq obj1, Eq obj2) => 
             (FRel e obj1 obj2 -> Bool) -> Lat e ->
             [obj1] -> [obj2] -> Lat (FRel e obj1 obj2)                               
fRelLatBy pr e s t = 
     Lat { lat_poSet   = fRelPoSetBy pr e s t
          ,lat_sup     = fJoin
          ,lat_inf     = fMeet
          ,lat_botEl   = fBot e s t
          ,lat_topEl   = fTop e s t
          ,lat_atomS   = filter pr (specialEls e (fBot e s t) $ lat_atomS e)
          ,lat_jIrredS = filter pr (specialEls e (fBot e s t) $ lat_jIrredS e)
          ,lat_mIrredS = filter pr (specialEls e (fTop e s t) $ lat_mIrredS e) }

fRelLat :: (Eq obj1, Eq obj2) =>
           Lat e -> [obj1] -> [obj2] -> Lat (FRel e obj1 obj2)
fRelLat = fRelLatBy $ const True

fRelLatList :: (Eq e, Eq obj1, Eq obj2) => 
               Lat e -> [FRel e obj1 obj2] -> Lat (FRel e obj1 obj2)
fRelLatList e rs = 
      let
        s = if null rs then [] else src $ head rs
        t = if null rs then [] else trg $ head rs
      in Lat { lat_poSet   = fRelPoSetList e rs
              ,lat_sup     = fJoin
              ,lat_inf     = fMeet
              ,lat_botEl   = fBot e s t
              ,lat_topEl   = fTop e s t
              ,lat_atomS   = specialEls e (fBot e s t) $ lat_atomS e
              ,lat_jIrredS = specialEls e (fBot e s t) $ lat_jIrredS e
              ,lat_mIrredS = specialEls e (fTop e s t) $ lat_mIrredS e }

\end{code}
