The next step is to provide data structures which respect the Haskell 98 standard. We %%@
then use them to parametrize our type classes so that the user is free to choose the %%@
variant he prefers. Furthermore, several consistency tests and standard operations shall %%@
be available within this module. 
\begin{code}
module Lattice where
import List ((\\),nub,intersect,union)
\end{code}
We start with the explanation of the provided data structures in accordance to the type %%@
classes. Again, the weakest structures are posets.
\begin{code}
data PoSet    el = PoSet { poSet_isElem   :: el -> Bool
                          ,poSet_elements :: [el]
                          ,poSet_lEq      :: el -> el -> Bool }
\end{code}
As shown here, we parametrize all record data types by the type of the elements. The %%@
functions are always analogous to the corresponding type class. 

A particularity of this approach can be seen with the next data structure. 
\begin{code}
data USemiLat el = USemiLat { uSemiLat_poSet    :: PoSet el
                             ,uSemiLat_sup      :: el -> el -> el
                             ,uSemiLat_topEl    :: el }
\end{code}
We do not have any inheritance as with the type classes. Thus, we have to carry the %%@
functions of the underlying poset by an explicit function (\smalltt{uSemiLat\_poSet}). To %%@
avoid longish names for accessing the poset functions, we (in analogy to RATH) introduce %%@
abbreviated notations. 
\begin{code}
uSemiLat_isElem   = poSet_isElem   . uSemiLat_poSet
uSemiLat_elements = poSet_elements . uSemiLat_poSet
uSemiLat_lEq      = poSet_lEq      . uSemiLat_poSet
\end{code}
The following data structures are constructed analogously.
\begin{code}
data LSemiLat el = LSemiLat { lSemiLat_poSet    :: PoSet el
                             ,lSemiLat_inf      :: el -> el -> el
                             ,lSemiLat_botEl    :: el
                             ,lSemiLat_atomS    :: [el] }

lSemiLat_isElem   = poSet_isElem   . lSemiLat_poSet
lSemiLat_elements = poSet_elements . lSemiLat_poSet
lSemiLat_lEq      = poSet_lEq      . lSemiLat_poSet

data RelCompLat el = RelCompLat { relCompLat_lSemiLat   :: LSemiLat el
                                 ,relCompLat_relComplem :: el -> el -> el  }

relCompLat_poSet    = lSemiLat_poSet    . relCompLat_lSemiLat
relCompLat_isElem   = lSemiLat_isElem   . relCompLat_lSemiLat
relCompLat_elements = lSemiLat_elements . relCompLat_lSemiLat
relCompLat_lEq      = lSemiLat_lEq      . relCompLat_lSemiLat
relCompLat_inf      = lSemiLat_inf      . relCompLat_lSemiLat
relCompLat_botEl    = lSemiLat_botEl    . relCompLat_lSemiLat
relCompLat_atomS    = lSemiLat_atomS    . relCompLat_lSemiLat 

data CompLat el = CompLat { compLat_lat     :: Lat el
                           ,compLat_complem :: el -> el }

compLat_isElem   = lat_isElem   . compLat_lat
compLat_elements = lat_elements . compLat_lat
compLat_lEq      = lat_lEq      . compLat_lat
compLat_sup      = lat_sup      . compLat_lat
compLat_inf      = lat_inf      . compLat_lat
compLat_topEl    = lat_topEl    . compLat_lat
compLat_botEl    = lat_botEl    . compLat_lat
compLat_atomS    = lat_atomS    . compLat_lat
compLat_jIrredS  = lat_jIrredS  . compLat_lat
compLat_mIrredS  = lat_mIrredS  . compLat_lat
compLat_poSet    = lat_poSet    . compLat_lat
\end{code}
The construction of standard lattices using upper and lower semi lattices, respectively, %%@
cannot be transferred that easily from the type classes. Obviously, we could get %%@
consistency problems because the data structures \smalltt{LSemiLat} and %%@
\smalltt{USemiLat} each carry their own poset. Therefore, we choose to redefine all %%@
functions explicitly. 
\begin{code}
data Lat el = Lat{ lat_poSet   :: PoSet el
                  ,lat_sup     :: el -> el -> el
                  ,lat_inf     :: el -> el -> el
                  ,lat_topEl   :: el
                  ,lat_botEl   :: el
                  ,lat_atomS   :: [el]
                  ,lat_jIrredS :: [el]
                  ,lat_mIrredS :: [el] }
\end{code}
Thus, the consistency problem is solved, but we have the new problem that the %%@
corresponding upper and lower semilattice, respectively, cannot be extracted that %%@
comfortably. Hence, we provide the functions 
\begin{code}
lat_uSemiLat :: Lat el -> USemiLat el
lat_uSemiLat l = USemiLat { uSemiLat_poSet = lat_poSet l
                           ,uSemiLat_sup   = lat_sup   l
                           ,uSemiLat_topEl = lat_topEl l }

lat_lSemiLat :: Lat el -> LSemiLat el                           
lat_lSemiLat l = LSemiLat { lSemiLat_poSet = lat_poSet l
                           ,lSemiLat_inf   = lat_inf   l
                           ,lSemiLat_botEl = lat_botEl l
                           ,lSemiLat_atomS = lat_atomS l }
\end{code}
for the user. Notice, that the structure \smalltt{Lat} carries the set of join and %%@
meet-irreducible elements in addition to the functions for upper and lower semilattices, %%@
respectively. 
With the abbreviations 
\begin{code}
lat_isElem   = poSet_isElem   . lat_poSet
lat_elements = poSet_elements . lat_poSet
lat_lEq      = poSet_lEq      . lat_poSet
\end{code}
the definition of our data structures is complete. Obviously, they directly correspond to 
the type class hierarchy of Figure \ref{LatticeTypeClasses} so that we are able to %%@
provide suitable instances later on.

The definition of the record data types above automatically provides the projections from %%@
stronger structures to weaker ones (e.g., from \smalltt{Lat} to \smalltt{LSemiLat}). But, %%@
there are also embeddings from the weak structures into stronger ones. In the following %%@
we provide standard implementations of these constructions. We first introduce some %%@
auxiliary functions.
\begin{code}
partialExt f x []     = x
partialExt f x (y:ys) = if f x y then partialExt f y ys
                        else partialExt f x ys

partialSupInf f els = \x y -> let bds = [e|e<-els,f x e,f y e]
                              in if null bds then x
                                 else partialExt (flip f) (head bds) (tail bds)

\end{code}
First of all, \smalltt{partialExt} determines the unique maximum/minimum of a given %%@
poset. With this, the parameter \smalltt{f} is the ordering on the elements given within %%@
the list \smalltt{y:ys}. If \smalltt{f} is given as $\leq$, the maximum is computed. If %%@
the reversed ordering $\geq$ is given, \smalltt{partialExt} determines the minimum.\\
The function \smalltt{partialSupInf} works analogously by computing the supremum/infimum %%@
of all elements of \smalltt{els} (if it exists) for the given ordering $\leq$/$\geq$. %%@
Thus, \smalltt{partialSupInf} already delivers two standard constructions and can be used %%@
by the following routines.
\begin{code}
poSet_uSemiLat :: PoSet el -> USemiLat el
poSet_uSemiLat p@(PoSet isElem els lEq) = 
  USemiLat { uSemiLat_poSet = p
            ,uSemiLat_sup   = partialSupInf lEq els
            ,uSemiLat_topEl = partialExt lEq (head els) (tail els) }

poSet_lSemiLat :: PoSet el -> LSemiLat el
poSet_lSemiLat p@(PoSet isElem els lEq) =
  let lSL = LSemiLat { 
               lSemiLat_poSet = p
              ,lSemiLat_inf   = partialSupInf (flip lEq) els
              ,lSemiLat_botEl = partialExt (flip lEq) (head els) (tail els)
              ,lSemiLat_atomS = atomSet lSL }
  in lSL

uSemiLat_lSemiLat = poSet_lSemiLat . uSemiLat_poSet 
lSemiLat_uSemiLat = poSet_uSemiLat . lSemiLat_poSet

lSemiLat_relCompLat :: LSemiLat el -> RelCompLat el
lSemiLat_relCompLat l@(LSemiLat p@(PoSet _ els lEq) inf _ _) = 
  RelCompLat { relCompLat_lSemiLat   = l
              ,relCompLat_relComplem = 
                  \x y -> let ok = [e|e<-els,lEq (inf x e) y]
                          in if null ok then y
                             else partialExt lEq (head ok) (tail ok) }

poSet_relCompLat :: PoSet el -> RelCompLat el
poSet_relCompLat = lSemiLat_relCompLat . poSet_lSemiLat

uSemiLat_lat :: Eq el => USemiLat el -> Lat el
uSemiLat_lat (USemiLat p@(PoSet _ els lEq) sup top) =
  let lat = Lat { lat_poSet   = p
                 ,lat_sup     = sup
                 ,lat_inf     = partialSupInf (flip lEq) els
                 ,lat_topEl   = top
                 ,lat_botEl   = partialExt (flip lEq) (head els) (tail els)
                 ,lat_atomS   = atomSet $ lat_lSemiLat lat
                 ,lat_jIrredS = jIrredSet lat
                 ,lat_mIrredS = mIrredSet lat }
  in lat

lSemiLat_lat :: Eq el => LSemiLat el -> Lat el
lSemiLat_lat (LSemiLat p@(PoSet _ els lEq) inf bot atomS) =
  let lat = Lat { lat_poSet   = p
                 ,lat_sup     = partialSupInf lEq els
                 ,lat_inf     = inf
                 ,lat_topEl   = partialExt lEq (head els) (tail els)
                 ,lat_botEl   = bot
                 ,lat_atomS   = atomS
                 ,lat_jIrredS = jIrredSet lat
                 ,lat_mIrredS = mIrredSet lat }
  in lat

poSet_lat :: Eq el => PoSet el -> Lat el
poSet_lat = uSemiLat_lat . poSet_uSemiLat

relCompLat_lat :: Eq el => RelCompLat el -> Lat el
relCompLat_lat = lSemiLat_lat . relCompLat_lSemiLat
\end{code}
The implementation is straightforward. The only new construction is provided within %%@
\smalltt{lSemiLat\_relCompLat}. It is a direct realization of Definition %%@
\ref{DefnRelCompl}.\\
Notice that these constructions strongly make use of the fact that we are dealing with %%@
finite, and, hence, complete structures. The correctness of the resulting data structures %%@
is no question that is to answer here. It has to be tested using the test routines %%@
introduced below.\\
Finally, we remain to provide embeddings into complementary lattices.
\begin{code}
lat_compLat :: Lat el -> CompLat el
lat_compLat l = 
  CompLat { compLat_lat     = l
           ,compLat_complem = flip (relCompLat_relComplem
                                   (lSemiLat_relCompLat $ lat_lSemiLat l)) $ 
                                   lat_botEl l }

poSet_compLat :: Eq el => PoSet el -> CompLat el
poSet_compLat = lat_compLat . poSet_lat

lSemiLat_compLat :: Eq el => LSemiLat el -> CompLat el
lSemiLat_compLat = lat_compLat . lSemiLat_lat

uSemiLat_compLat :: Eq el => USemiLat el -> CompLat el
uSemiLat_compLat = lat_compLat . uSemiLat_lat

relCompLat_compLat :: Eq el => RelCompLat el -> CompLat el
relCompLat_compLat l@(RelCompLat lSL rc) = 
  CompLat { compLat_lat     = relCompLat_lat l
           ,compLat_complem = flip rc $ lSemiLat_botEl lSL }
\end{code}
Again, the implementation is straightforward. The embedding from \smalltt{RelCompLat} %%@
into complementary lattices is trivial since we already have a function for determining %%@
relative pseudo complements. With all other functions, we have to use the standard %%@
construction from above.

Thus, we are ready to switch to operations on lattices and the test functions. As we are %%@
only dealing with complete lattices, fixpoint operations become important. From Theorems %%@
\ref{Tarski1} and \ref{Tarski2} we know that the fixpoints of a monotone function %%@
$f:\SL\lora\SL$ over a complete lattice $\SL$ form a sublattice of $\SL$. Furthermore, we %%@
know that $f$ has at least one fixed point. The applications using this fact are %%@
manyfold. Therefore, we provide the following functions. 
\begin{code}
leastFPA ::  Eq a => a -> (Lat a) -> (a -> a) -> a 
leastFPA a l f = if lat_lEq l a (f a) then leastFPA' a
                 else leastFPA' $ foldl1 (lat_inf l) [x|x<-lat_elements l,
                                                        lat_lEq l a (f x)]
  where
    leastFPA' b = if (f b) == b then b
                  else leastFPA' (f b)

leastFP ::  Eq a => (Lat a) -> (a -> a) -> a 
leastFP l = leastFPA (lat_botEl l) l

latFPs :: Eq a => (Lat a) -> (a -> a) -> (Lat a)
latFPs l f = let els = [x | x <- lat_elements l, f x == x ]
             in Lat { lat_poSet = PoSet { poSet_isElem   = flip elem $ els
                                         ,poSet_elements = els
                                         ,poSet_lEq      = lat_lEq l }
                     ,lat_sup     = lat_sup l
                     ,lat_inf     = lat_inf l
                     ,lat_botEl   = foldl (lat_inf l) (lat_topEl l) els
                     ,lat_topEl   = foldl (lat_sup l) (lat_botEl l) els
                     ,lat_atomS   = atomSetBy (lat_poSet l) (lat_botEl l) 
                                              (lat_inf l) []
                     ,lat_jIrredS = irredSetBy (lat_poSet l) (lat_botEl l) 
                                               (lat_sup l) []
                     ,lat_mIrredS = irredSetBy (lat_poSet l) (lat_topEl l) 
                                               (lat_inf l) [] }
\end{code}
First of all, \smalltt{leastFPA} determines the least fixpoint of the monotone function %%@
\smalltt{f}, which is greater or equal to a given element \smalltt{a} from the underlying %%@
lattice \smalltt{l}. Furthermore, \smalltt{leastFP} delivers the standard variant by %%@
computing the least fixpoint of \smalltt{f}. The lattice of all fixpoints of \smalltt{f} %%@
is generated by \smalltt{latFPs}. Notice that we cannot rely on the standard %%@
\smalltt{Prelude} functions \smalltt{minimum} resp.\ \smalltt{maximum} to determine the %%@
least resp.\ greatest element within the list of fixed points since these elements do not %%@
necessarily exist. Obviously, the correct termination of all these functions above %%@
strongly depends on whether \smalltt{f} is monotonic. This property can be tested by the %%@
function \smalltt{testMonoFunc} introduced later on. The routines for determining the %%@
atoms and irreducible elements, respectively, are also explained later.

Now, our attention goes to the consistency tests. As we are dealing with potentially big %%@
lattices, some special thoughts go to efficiency considerations. As mentioned above, in a %%@
finite lattice every element can be computed (or partitioned) by a set of irreducible %%@
elements. In a distributive lattice this representation is even unique. Atoms play an %%@
analogous role within Boolean lattices. The next lemma shows that all tests in a %%@
standard/Boolean lattice can be reduced to the irreducible elements/atoms. 
\begin{lem}
\label{LemReduceTests}
Let $\SL$ be a set. and $\vee : \SL\times\SL\lora\SL$ and its extension $\bigvee$ be an %%@
operation on $\SL\times\SL$. If there is an inclusionminimal subset $S$ of $\SL$ such %%@
that every element $x\in\SL$ can be computed by $\bigvee S'$ for a subset $S'$ of $S$, we %%@
have : 
\begin{tabbing}
\qquad \= (1) \= If $\vee$ reduced to $S$ is commutative/associative/idempotent then\\
       \>     \> it is commutative/associative/idempotent in $\SL$.\\
       \> (2) \> If $\vee$ reduced to $S$ is commutative and idempotent then\\
       \>     \> it is idempotent in $\SL$.
\end{tabbing}
\end{lem}
The proof is obvious and, therefore, omitted.

Since atoms are join-irreducible elements and in an atomic lattice these two notions are %%@
even equivalent, it would suffice to provide a function that computes the irreducible %%@
elements. But, atoms are defined via the meet operation and join-irreducible elements via %%@
the join operation. One of these operations can be evidently more inefficient than the %%@
other. 

Thus, we provide both --- standard functions to compute the atoms and irreducible %%@
elements, respectively. We start with the function \smalltt{atomSetBy}. 
\begin{code}
atomSetBy :: PoSet el -> el -> (el -> el -> el) -> [el] -> [el] 
atomSetBy p botEl inf els = 
  let 
    li = if null els then atomSet' [] (poSet_elements p) 0 
         else atomSet' [] els 0 
  in [head el | el <- map (\x -> filter (\y -> (poSet_lEq p) y x) li) li, 
      null $ tail el] 
  where 
    atomSet' at []  n = if n>0 then atomSet' [] at 0 else at 
    atomSet' at [a] n = if n>0 then atomSet' [] (a:at) 0 else (a:at) 
    atomSet' at (a1:(a2:els)) n 
               | (poSet_lEq p) a1 botEl          = atomSet' at (a2:els) (n+1) 
               | (poSet_lEq p) a2 botEl          = atomSet' at (a1:els) (n+1) 
               | (poSet_lEq p) a1 a2             = atomSet' (a1:at) els (n+1) 
               | (poSet_lEq p) a2 a1             = atomSet' (a2:at) els (n+1) 
               | (poSet_lEq p) (inf a1 a2) botEl = atomSet' (a1:(a2:at)) els n 
               | otherwise                       = atomSet' at els (n+1) 
\end{code}
It is parametrized by the underlying poset \smalltt{p}, the least element %%@
\smalltt{botEl}, the meet operation \smalltt{inf} and a list \smalltt{els} of elements to %%@
which the computation shall be reduced (e.g., a sublattice). If \smalltt{els} is empty, %%@
all atoms of the underlying poset are computed.\\
The parametrization comes due to the intent to make it comfortable for the user to %%@
compute the atoms directly when he creates the data structure for the corresponding %%@
(lower semi) lattice. Thus, a lower semilattice for the divisibility relation could be %%@
created by
\begin{smallverb}
divLat = let pS = PoSet { poSet_isElem   = flip elem $ [1,2,3,4,6,12]
                         ,poSet_elements = [1,2,3,4,6,12]
                         ,poSet_lEq      = \x y -> mod y x == 0 }
         in LSemiLat { lSemiLat_poSet = pS
                      ,lSemiLat_inf   = gcd
                      ,lSemiLat_botEl = 1
                      ,lSemiLat_atomS = atomSetBy pS 1 gcd [] }       
\end{smallverb}
Alternatively, the user can assign the empty list to \smalltt{lSemiLat\_atomS}. For this %%@
case, we provide the standard function 
\begin{code}
atomSet :: LSemiLat el -> [el]
atomSet l = atomSetBy (lSemiLat_poSet l) (lSemiLat_botEl l) (lSemiLat_inf l) []
\end{code}
which can be used later on to compute the atoms ``on demand''. As shown here, the %%@
extended parametrized variant of any function will be ended by \smalltt{By} throughout %%@
the module.\\
Back to the function above, the atoms are computed by filtering out those elements $\neq %%@
0$ that have exactly one element in the lattice that is lower or equal to them. This %%@
computation tends to be rather extensive. Hence, we reduce the list of possible atoms %%@
before we filter them out. This is done by calling \smalltt{atomSet'} which establishes %%@
the key functionality. It successively constructs the list of atoms \smalltt{at} by local %%@
comparison of the first two elements in the element list. The element list steadily %%@
becomes shorter. To detect when the element list has been traversed without causing any %%@
changes in the atom list, we use the parameter \smalltt{n}. If it is zero and the list of %%@
remaining elements is empty, the computation is completed. Obviously, the result of %%@
\smalltt{atomSet'} strongly depends on the arrangement of the elements within the element %%@
list. Thus, it does not necessarily deliver the set of atoms so that they have to be %%@
filtered out after \smalltt{atomSet'} is finished. Consider the following example %%@
computation for the lattice induced by the powerset $\PP(\{ 1,2,3 \})$. If we, for %%@
example, have the element list \smalltt{[[1],[1,2],[2],[2,3],[3],[1,3],[],[1,2,3]]} and %%@
the underlying poset \smalltt{pS}, calling \smalltt{atomSetBy pS [] intersect [] } %%@
results in 
\begin{tabbing}
\quad \= \smalltt{atomSet'} \= \smalltt{[]}\qquad\qquad\qquad \= 
         \smalltt{[[1],[1,2],[2],[2,3],[3],[1,3],[],[1,2,3]]} \= \smalltt{0} \\
      \> \smalltt{atomSet'} \> \smalltt{[[1]]}          \> 
         \smalltt{[[2],[2,3],[3],[1,3],[],[1,2,3]]}           \> \smalltt{1}\quad \= %%@
(guard 3  )\\
      \> \smalltt{atomSet'} \> \smalltt{[[1],[2]]}      \> 
         \smalltt{[[3],[1,3],[],[1,2,3]]}                     \> \smalltt{2} \> (guard 3  %%@
)\\
      \> \smalltt{atomSet'} \> \smalltt{[[1],[2],[3]]}  \> 
         \smalltt{[[],[1,2,3]]}                               \> \smalltt{3} \> (guard 3  %%@
)\\
      \> \smalltt{atomSet'} \> \smalltt{[[1],[2],[3]]}  \> 
         \smalltt{[[1,2,3]]}                                  \> \smalltt{4} \> (guard 1  %%@
)\\
      \> \smalltt{atomSet'} \> \smalltt{[]}             \> 
         \smalltt{[[1,2,3][1],[2],[3]]}                       \> \smalltt{0} \> (since %%@
n>0)\\
      \> \smalltt{atomSet'} \> \smalltt{[[1]]}          \> 
         \smalltt{[[2],[3]]}                                  \> \smalltt{1} \> (guard 4  %%@
)\\
      \> \smalltt{atomSet'} \> \smalltt{[[1],[2],[3]]}  \> 
         \smalltt{[]}                                         \> \smalltt{1} \> (guard 5  %%@
).
\end{tabbing}
Since \smalltt{n} is $1$, the list is again traversed but no changes take place so that %%@
we omit these function calls. In this example, we use an arrangement of the elements such %%@
that \smalltt{atomSet'} does indeed deliver the atoms. But, consider the list %%@
\smalltt{[[1],[2,3],[2],[1,3],[3],[1,2],[],[1,2,3]]}. Obviously, here only the least %%@
element \smalltt{[]} is taken out. Thus, the final filtering in \smalltt{atomSet} is %%@
necessary.

Now, we switch to the determination of the irreducible elements. Since their definition %%@
is very inefficient for computational purposes, we take the detour over the reducible %%@
elements.
\begin{code}
redSetBy :: (Eq el) => PoSet el -> (el -> el -> el) -> [el] -> [el]
redSetBy p sup_inf els = 
  let li = if null els then poSet_elements p
           else els
  in nub $ redSet' [] (takeN 2 li)
  where
    redSet' red []     = red
    redSet' red (x:xs) = let hX   = head x
                             lX   = last x
                             s    = sup_inf hX lX
                         in if s==hX || s==lX then redSet' red xs
                                              else redSet' (s : red) xs
\end{code}
The parametrization is analogous to \smalltt{atomSet}, but we here have the opportunity %%@
to compute either the join reducible or the meet reducible elements. Thus, we call the %%@
corresponding parameter \smalltt{sup\_inf}. First, we compute all pairs of two elements %%@
by means of the function \smalltt{takeN} and then call \smalltt{redSet'}.
\begin{code}
takeN :: Int -> [a] -> [[a]]
takeN _ []       = []
takeN 1 list     = [[x] | x<-list]
takeN n (l:list) = [l:li | li<-takeN (n-1) list]++takeN n list
\end{code}
Here, the sup/inf of any pair of elements is computed and compared to these elements. If %%@
it is not equal to any of these two elements, it is reducible (to these elements). %%@
Notice, that \smalltt{takeN} 
can only be used for commutative operations since it only delivers pairs of elements %%@
without considering their arrangement. Calling \smalltt{takeN 2 [1,2,3]}, for instance, %%@
results in the list \smalltt{[[1,2],[1,3],[2,3]]}. If we have a non-commutative operation %%@
that shall be applied to {\em all} pairs of elements, we thus have to use list %%@
comprehension or equivalent constructs.

With the reducible elements we automatically have the irreducible elements.
\begin{code}
irredSetBy :: (Eq el) => PoSet el -> el -> (el -> el -> el) -> [el] -> [el]
irredSetBy p bot_topEl sup_inf els = let li = if null els then poSet_elements p
                                              else els
                                     in li \\ (bot_topEl:(redSetBy p sup_inf li))
\end{code}
The only thing we have to mind is to take out the least/greatest element since it is not %%@
join/meet-irreducible by definition. Notice that we need the least element and the join %%@
operation to compute the join-irreducible elements and the greatest element and the meet %%@
operation to compute the meet-irreducible elements. \\
Again, we have standard versions for calling the functions above.
\begin{code}
jRedSet,mRedSet,jIrredSet,mIrredSet :: (Eq el) => Lat el -> [el]
jRedSet l   = redSetBy   (lat_poSet l) (lat_sup l) []
mRedSet l   = redSetBy   (lat_poSet l) (lat_inf l) []

jIrredSet l = irredSetBy (lat_poSet l) (lat_botEl l) (lat_sup l) []
mIrredSet l = irredSetBy (lat_poSet l) (lat_topEl l) (lat_inf l) []
\end{code}
Notice that the determination of the atoms and irreducible elements, respectively, causes %%@
quadratic runtime costs in the worst case scenario. With \smalltt{atomSet} it is possible %%@
that only one element is taken out with every pass through the element list. With %%@
\smalltt{redSet} we have to generate all pairs of elements, which naturally results in %%@
quadratic computation times. But, consider that the average case for \smalltt{atomSet} %%@
should be evidently better than for \smalltt{irredSetBy} / \smalltt{redSetBy} (provided %%@
that join and meet have comparable computation times). This comes due to the fact that %%@
with \smalltt{irredSetBy} the worst case scenario is created a priori by computing all %%@
pairs of elements.

Now, we can proceed with the test functions. Before we start, we want to mention that the %%@
following data structures are already defined within the RATH module \smalltt{RelAlg} to %%@
test category
theoretical properties. We redefine them so that we have both --- conformance to RATH and %%@
a stand alone module \smalltt{Lattice}. 

Of course, we want to know when a test failed and why it failed. Thus, the result of a 
failed test can be represented by the data structure 
\begin{code}
type Result el = (String,[el]) 
\end{code}
which holds the reason for the failure and a list of the affected elements. From the %%@
definitions of the lattice data structures, extensive list operations (in particular, the %%@
concatenation operator \smalltt{(++)} to combine several tests) are to expect. As lists %%@
are defined recursively, their concatenation is very slow. In accordance to RATH, we use %%@
the standard escape. We ``introduce'' a new concatenation operator. Consider the lists %%@
\smalltt{l1=["A","B"]} and \smalltt{l2=["C","D"]}. If we represent them by the functions 
\begin{smallverb}
l1,l2 :: [String] -> [String] 
l1 = let 
       f1 = ("A":) 
       f2 = ("B":) 
     in f1 . f2 
l2 = let 
       f3 = ("C":) 
       f4 = ("D":) 
     in f3 . f4 
\end{smallverb}
we have the composition operator \smalltt{(.)} as our new list concatenation. The %%@
resulting list \smalltt{l1++l2} then can be computed by \smalltt{(l1.l2) []}. This %%@
motivates the definition of 
\begin{code}
type TestRes el = [Result el] -> [Result el] 
\end{code}
which represents a list of failed tests. Furthermore, we need to compute a list of test %%@
results later on. This can be done by extending the predefined function \smalltt{foldr} %%@
to \smalltt{foldF} (fold function) 
\begin{code}
foldF :: [a -> a] -> a -> a 
foldF l r = foldr id r l 
\end{code}
which we here use at the type \smalltt{[TestRes el] -> TestRes el}. Last but not least, %%@
we provide a function \smalltt{testP} (test predicate) 
\begin{code}
testP :: Bool -> [el] -> String -> TestRes el 
testP p els s = \res -> if p then res else (s,els) : res 
\end{code}
to generate a parametrized test result, if the predicate \smalltt{p} is \smalltt{False}. %%@
Thus, we have all necessary data structures so that we now can proceed with the single %%@
consistency tests. 

First, the question arises which tests are needed and how to parametrize them. Because of %%@
the two approaches to lattices, we provide order theoretical tests as well as algebraic %%@
ones. The three functions
\begin{code}
testRefl, testTrans :: PoSet el -> TestRes el
testRefl     p  = foldF [testP (poSet_lEq p o o) [o] "Reflexivity:" 
                         | o <- poSet_elements p]

testTrans    p  = let incl = poSet_lEq p 
                      els  = poSet_elements p
                  in foldF [testP (not (incl x1 x2 && incl x2 x3) || incl x1 x3) 
                                  [x1,x2,x3] "Transitivity:"
                            | x1 <- els, x2 <- els, x3<-els]
                            
testAntiSymm :: (Eq el) => PoSet el -> TestRes el
testAntiSymm p  = let incl = poSet_lEq p
                  in foldF [let x1 = head els
                                x2 = last els
                            in testP (not (incl x1 x2 && incl x2 x1) || x1==x2) 
                                     [x1,x2] "Antisymmetry:"
                            | els <- takeN 2 $ poSet_elements p]
\end{code}
together with the all-in-one test
\begin{code}
testPoSet :: (Eq el) => PoSet el -> TestRes el
testPoSet p = testRefl p . testTrans p . testAntiSymm p
\end{code}
constitute the order theoretical part by testing reflexivity, transitivity and %%@
antisymmetry of a poset. Notice that \smalltt{testTrans} does not use \smalltt{takeN} %%@
because the arrangement of the three elements in question is essential. In contrast, %%@
\smalltt{testAntiSymm} can be supplied by \smalltt{takeN}.

The algebraic tests are quite more extensive. Because of the symmetric definition of %%@
(algebraic) lattices, we use an at first sight somehow unnatural parametrization. Let us %%@
have a look at the first test, which verifies that a poset is closed under a binary %%@
operation (e.g. join, meet, relative pseudo complement).
\begin{code}
testConsBy :: (Eq el) => PoSet el -> (el -> el -> el) -> [el] -> [el] -> 
                         TestRes el   
testConsBy p op els targetEls = 
  foldF [testP ((flip elem $ targetEls)$op (head xs) (last xs)) xs "Consistency:" 
         |  xs<-takeN 2 $ els]
\end{code}
The function first of all needs the underlying poset \smalltt{p}. Furthermore, the %%@
operation \smalltt{op} that shall be tested is necessary. The third parameter %%@
\smalltt{els} can be used to reduce the test to the elements given in it, i.e., to set %%@
the domain of \smalltt{op}. In contrast to \smalltt{atomSetBy} and \smalltt{irredSetBy}, %%@
\smalltt{els} must not be empty. The tests then are not automatically applied to all %%@
elements of the underlying poset. We prefer this variant to avoid longish \smalltt{if ... %%@
then} phrases. Furthermore, we provide a standard variant for each test so that this %%@
approach should be acceptable. But, notice that we have to use the \smalltt{elem} %%@
function to determine whether an element is in \smalltt{els}. This can cause a loss in %%@
efficiency when membership within the lattice is decided by a certain predicate and not %%@
by checking if it is in the element list. Thus, it is up to the user to decide whether %%@
\smalltt{testConsBy} can be used efficiently.\\
The last parameter can be used to set the range of \smalltt{op}. Again, %%@
\smalltt{targetEls} has no default setting if an empty list is given as parameter. The %%@
following three examples show how \smalltt{testConsBy} can be used to achieve certain %%@
goals.
\begin{enumerate}
\item If we want to test the join consistency of a lattice \smalltt{l}, we can reduce the %%@
test to the join-irreducible elements whereas the range of the join operation is the set %%@
of all elements of the lattice. Thus, we can use the function above by\\
\centerline{\smalltt{testConsBy (lat\_poSet l) (lat\_sup l) (lat\_jIrredS l) %%@
(lat\_elements l)}.}
\item Suppose we want to test whether some elements form a sublattice of an arbitrary %%@
lattice \smalltt{l} with respect to the join operation. Then we can use %%@
\smalltt{testConsBy} by\\
\centerline{\smalltt{testConsBy (lat\_poSet l) (lat\_sup l) ({\em<elements>}) %%@
({\em<elements>})}.}
\item To test whether a lattice \smalltt{l} is closed under join without reducing the %%@
test to any elements, one has to use the default call\\
\centerline{\smalltt{testConsBy (lat\_poSet l) (lat\_sup l) (lat\_elements l) %%@
(lat\_elements l)}.}
\end{enumerate}
Consider that, if the order theoretic tests as well as the two consistency tests above %%@
succeed for join and meet, all remaining algebraic tests for standard lattices are %%@
redundant. Nearly the same is true for the opposite. If all algebraic tests (even without %%@
the join/meet consistency) succeed, the induced order is indeed an order and, hence, the %%@
order theoretic tests need not be performed.\\
To make the use of the test functions more comfortable, we introduce standard calls to %%@
check join/ meet consistency as well as consistency with the relative pseudo complements.
\begin{code}
testJCons :: (Eq el) => USemiLat el -> TestRes el
testJCons l = testConsBy (uSemiLat_poSet l)    (uSemiLat_sup l) 
                         (uSemiLat_elements l) (uSemiLat_elements l)

testMCons :: (Eq el) => LSemiLat el -> TestRes el
testMCons l = testConsBy (lSemiLat_poSet l)    (lSemiLat_inf l)
                         (lSemiLat_elements l) (lSemiLat_elements l)

testRelCompCons :: (Eq el) => RelCompLat el -> TestRes el
testRelCompCons l = testConsBy (relCompLat_poSet l)    (relCompLat_relComplem l)
                               (relCompLat_elements l) (relCompLat_elements l)
\end{code}
The complement is a unary operation, so that \smalltt{testConsBy} cannot be used. Thus, %%@
we introduce a second variant together with a standard call.
\begin{code}
testConsUnBy :: (Eq el) => PoSet el -> (el -> el) -> [el] -> [el] -> TestRes el
testConsUnBy p op els targetEls = 
    foldF [testP ((flip elem) targetEls $ op x) [x] "ConsistencyUn:" 
           | x<-els]

testCompCons :: (Eq el) => CompLat el -> TestRes el
testCompCons l = testConsUnBy (compLat_poSet l)    (compLat_complem l)
                              (compLat_elements l) (compLat_elements l)
\end{code}

Commutativity, associativity, idempotency and the absorption laws of the join/meet %%@
operation can be verified by the following test routines.
\begin{code}
testCommBy,testIdemBy, testAssBy :: (Eq el) => 
                              PoSet el -> (el -> el -> el) -> [el] -> TestRes el
testCommBy p op els = foldF [let o1   = head os
                                 o2   = last os
                             in testP (op o1 o2 == op o2 o1) os "Commutativity:" 
                             | os<-takeN 2 els]

testIdemBy p op els = foldF [testP (op o o == o) [o] "Idempotency:" 
                             | o<-els]
                                                                               
testAssBy p op els = foldF [let o1    = head os
                                o2    = head $ tail os
                                o3    = last os
                            in testP ((op o1 $ op o2 o3) == (op (op o1 o2) o3)) 
                                     os "Associativity:" 
                            | os<-takeN 3 els]
testAbsBy :: (Eq el) => PoSet el -> (el -> el -> el) ->
                                    (el -> el -> el) -> [el] -> TestRes el
testAbsBy p op1 op2 els = foldF [testP ((op1 o1 $ op2 o1 o2)==o1) [o1,o2] 
                                 "Absorption:" | o1<-els, o2<-els]
\end{code}
They are parametrized analogous to \smalltt{testConsBy}. Obviously, the list %%@
\smalltt{targetEls} can be left out since it plays no role for these tests. Notice that %%@
\smalltt{testAbsBy} cannot rely on \smalltt{takeN}. Again, the standard variants are %%@
provided by the following functions.
\begin{code}
testJComm,testJIdem, testJAss :: (Eq el) => USemiLat el -> TestRes el
testJComm l= testCommBy (uSemiLat_poSet l) (uSemiLat_sup l) (uSemiLat_elements l)
testJIdem l= testIdemBy (uSemiLat_poSet l) (uSemiLat_sup l) (uSemiLat_elements l)
testJAss  l= testAssBy  (uSemiLat_poSet l) (uSemiLat_sup l) (uSemiLat_elements l)

testMComm,testMIdem, testMAss :: (Eq el) => LSemiLat el -> TestRes el
testMComm l= testCommBy (lSemiLat_poSet l) (lSemiLat_inf l) (lSemiLat_elements l)
testMIdem l= testIdemBy (lSemiLat_poSet l) (lSemiLat_inf l) (lSemiLat_elements l)
testMAss  l= testAssBy  (lSemiLat_poSet l) (lSemiLat_inf l) (lSemiLat_elements l)

testJAbs,testMAbs :: (Eq el) => Lat el -> TestRes el
testJAbs  l= testAbsBy (lat_poSet l) (lat_sup l) (lat_inf l) (lat_elements l)
testMAbs  l= testAbsBy (lat_poSet l) (lat_inf l) (lat_sup l) (lat_elements l)
\end{code}
Furthermore, we need two functions to verify the correctness of the greatest and least %%@
element, respectively.
\begin{code}
testTopElBy :: USemiLat el -> [el] -> TestRes el
testTopElBy l els =
  let top = uSemiLat_topEl l
  in foldF ( testP (uSemiLat_isElem l top) [top] "Top in List:" :
            [testP (uSemiLat_lEq l x top)  [x] "Top element:" | x<- els] )

testBotElBy :: LSemiLat el -> [el] -> TestRes el
testBotElBy l els = 
  let bot = lSemiLat_botEl l
  in foldF ( testP (lSemiLat_isElem l bot) [bot] "Bot in List:" :
            [testP (lSemiLat_lEq l bot x)  [x] "Bottom element:" | x<- els] )
                                 
testTopEl :: USemiLat el -> TestRes el
testTopEl l = testTopElBy l (uSemiLat_elements l)

testBotEl :: LSemiLat el -> TestRes el
testBotEl l = testBotElBy l (lSemiLat_elements l)
\end{code}
Hence, we have accomplished the pure consistency tests for standard (upper/lower %%@
semi)lattices. Now, we remain to provide all-in-one tests. This is done by the following %%@
functions.
\begin{code}
testUpSemiLatticeBy :: (Eq el) => USemiLat el -> [el] -> [el] -> TestRes el
testUpSemiLatticeBy  l els targetEls = 
      let pS  = uSemiLat_poSet l
          s   = uSemiLat_sup   l
      in testCommBy pS s els . testAssBy  pS s els . 
         testIdemBy pS s els . testConsBy pS s els targetEls .
         testTopElBy l els

testLoSemiLatticeBy :: (Eq el) => LSemiLat el -> [el] -> [el] -> TestRes el
testLoSemiLatticeBy  l els targetEls = 
      let pS = lSemiLat_poSet l
          i  = lSemiLat_inf   l
      in testCommBy pS i els . testAssBy  pS i els .
         testIdemBy pS i els . testConsBy pS i els targetEls .
         testBotElBy  l els
                                      
testLatticeBy       :: (Eq el) => Lat el -> [el] -> [el] -> 
                                            [el] -> [el] -> TestRes el                                       
testLatticeBy        l els1 els2 els3 targetEls = 
         testUpSemiLatticeBy (lat_uSemiLat l) els1 targetEls  . 
         testLoSemiLatticeBy (lat_lSemiLat l) els2 targetEls  .
         testAbsBy (lat_poSet l) (lat_sup l) (lat_inf l) els3 .
         testAbsBy (lat_poSet l) (lat_inf l) (lat_sup l) els3
\end{code}
It should be clear from the previous explanations how the first two functions work. The %%@
test for lattices is special in the sense that we have three parameters \smalltt{els1}, %%@
\smalltt{els2} and \smalltt{els3}, respectively, to reduce the tests for upper and lower %%@
semilattices. Good examples for \smalltt{els1} and \smalltt{els2} are the join resp.\ %%@
meet-irreducible elements. Notice that it makes no sense to differentiate the parameter %%@
\smalltt{targetEls} since we want to test a lattice structure.
 
The standard use of the function above is the following.
\begin{code}
testUpSemiLattice :: (Eq el) => USemiLat el -> TestRes el
testUpSemiLattice l = testUpSemiLatticeBy l 
                          (uSemiLat_elements l) (uSemiLat_elements l)

testLoSemiLattice :: (Eq el) => LSemiLat el -> TestRes el
testLoSemiLattice l = testLoSemiLatticeBy l  
                          (lSemiLat_elements l) (lSemiLat_elements l)

testLattice       :: (Eq el) => Lat el -> TestRes el
testLattice       l = testLatticeBy l (lat_jIrredS l)  (lat_mIrredS l)
                                      (lat_elements l) (lat_elements l)
\end{code}
Notice that \smalltt{testLattice} automatically reduces the test of the upper resp.\ %%@
lower semilattice to the join resp.\ meet-irreducible elements. Of course, this %%@
presupposes that the lists are not empty.

The next step is to check modularity, distributivity and atomicity. We start with %%@
modularity.
\begin{code}
testModularBy :: (Eq el) => Lat el -> [el] -> TestRes el
testModularBy l els = 
   let sup = lat_sup l
       inf = lat_inf l
   in foldF [testP (not (lat_lEq l x1 x3) || (sup x1 $ inf x2 x3) == 
                   inf (sup x1 x2) x3) [x1,x2,x3] "Modularity:"
             | x1<-els, x2<-els, x3<-els]

testModular :: (Eq el) => Lat el -> TestRes el
testModular l = testModularBy l (lat_elements l)
\end{code}
From Lemma \ref{LemEqualDistrLaw} we know that the validities of the modular laws imply %%@
each other. Thus, we directly parametrize the function above by the underlying lattice %%@
\smalltt{l}. Notice that we cannot use \smalltt{takeN} to determine all combinations of %%@
any three elements out of \smalltt{l}, because the arrangement of the three elements in %%@
question is essential (cf.\ Definition \ref{DefnModLat}). Furthermore, the modularity %%@
test can, in general, not be reduced to the atoms or join-irreducible elements, %%@
respectively.

The same is true for the following routine which can be used to check distributivity.
\begin{code}
testDistrBy :: (Eq el) => PoSet el -> (el -> el -> el) -> 
               (el -> el -> el) -> [el] -> TestRes el
testDistrBy p op1 op2 els =
    foldF [let x1   = head xs
               x2   = xs!!1
               x3   = last xs
           in testP ((op1 x1 $ op2 x2 x3) == op2 (op1 x1 x2) (op1 x1 x3)) 
                    xs "Distributivity:"
           | xs<-takeN 3  els]
\end{code}
Again, the validity of one of the distributive laws implies the validity of the other %%@
(cf.\ Lemma \ref{LemEqualDistrLaw}) so that the parametrization of this function by the %%@
two operations \smalltt{op1} and \smalltt{op2} is redundant from a mathematical point of %%@
view. But, again  efficiency considerations with respect to the join and meet operations %%@
drove us to this decision.
To determine join distributivity, we use the join operation three times and the meet %%@
operation only twice. The opposite is true for meet distributivity. The user now can %%@
decide which variant is the most effective. Hence, we provide standard tests for join and %%@
meet distributivity, respectively.
\begin{code}
testJDistr,testMDistr :: (Eq el) => Lat el -> TestRes el
testJDistr l = testDistrBy (lat_poSet l) (lat_sup l) (lat_inf l) (lat_elements l)

testMDistr l = testDistrBy (lat_poSet l) (lat_inf l) (lat_sup l) (lat_elements l)
\end{code}
From Theorem \ref{ThmPseudoDistr} we know that the notions of complete upwards %%@
distributivity and complete Brouwerian lattices are equivalent. Furthermore, we can %%@
conclude the equivalence of complete upwards and downwards distributivity within finite %%@
lattices. Thus, testing distributivity can be substituted by testing the existence of the %%@
relative pseudo complements for all pairs of elements and vice versa.
\begin{code}
testRelComplBy :: (Eq el) => RelCompLat el -> [el] -> TestRes el
testRelComplBy l els = 
    foldF [testP ((flip elem) els (relCompLat_relComplem l 
                          (head xs) (last xs))) xs "Rel.pseudo compl.:" 
           | xs <- takeN 2 els ]

testRelCompl :: (Eq el) => RelCompLat el -> TestRes el
testRelCompl l = testRelComplBy l (relCompLat_elements l)
\end{code}
Now, we proceed with atomicity. This test tends to be rather costly with respect to %%@
efficiency. Unfortunately, we have no general possibility to delimit it. One approach %%@
could be to test whether all join-irreducible elements are atoms. But, the quadratic %%@
complexity of determining the join-irreducible elements indicated by \smalltt{irredSetBy} %%@
only permits a gain in efficiency, if the user is able to provide a faster way to compute %%@
them. Thus, we deliver two functions --- one that tests the join-irreducible elements and %%@
one that directly checks whether all elements can be computed by the disjunction of a %%@
certain number of atoms.
\begin{code}
testAtomicIrred :: (Eq el) => Lat el -> TestRes el
testAtomicIrred l = 
    let incl   = lat_lEq l
    in foldF [ let x1 = head xs
                   x2 = last xs
               in testP (not (incl x1 x2 || incl x2 x1 ) || x1==x2)
                        xs "Atomicity(irr.el.):"  | xs<-takeN 2 $ lat_jIrredS l]

testAtomicBy :: (Eq el) => Lat el -> [el] -> TestRes el
testAtomicBy l els = 
  foldF [let as  = [ a  | a<-lat_atomS l, (lat_lEq l) a x ]
         in if null as then id
            else testP (x==foldl1 (lat_sup l) as) (x:as) "Atomicity:"| x<-els]

testAtomic :: (Eq el) => Lat el -> TestRes el
testAtomic l = testAtomicBy l (lat_elements l)
\end{code}
The parametrization of these two functions is a bit different. If the user, for example, %%@
wants to test a sublattice of \smalltt{l}, he has to create a new data structure or to %%@
use \smalltt{testAtomicBy}.\\
The atomicity test within \smalltt{testAtomicIrred} is done by checking whether there are %%@
two (different) elements such that they are in the underlying order relation. If not, all %%@
join-irreducible elements are atoms.\\
Notice that we leave the user to check whether the atom set and set of irreducible %%@
elements, respectively, are correct. This can, for example, be done by comparing them to %%@
the result of \smalltt{atomSetBy} and \smalltt{irredSetBy}, respectively.

Now, only the complement operation is left. The check is done by the following functions.
\begin{code}
testComplBy :: CompLat el -> [el] -> TestRes el 
testComplBy l els = 
  let lEq     = compLat_lEq l
      complem = compLat_complem l
      in foldF [testP (lEq (compLat_topEl l) (compLat_sup l x (complem x)) &&
                       lEq (compLat_inf l x $ complem x) (compLat_botEl l) &&
                       lEq x (complem $ complem x)                         && 
                       lEq (complem $ complem x) x   ) [x] "Complement:" 
                | x <- els]

testCompl :: CompLat el -> TestRes el 
testCompl l = testComplBy l (compLat_elements l)
\end{code}
Obviously, we test whether $x\vee\ol{x}=1$, $x\wedge\ol{x}=0$ and $\ol{\ol{x}}=x$ hold.

Thus, we can introduce all-in-one tests for Boolean lattices.
\begin{code}
testBoolLatticeBy :: (Eq el)=>CompLat el->[el]->[el]->[el]->[el]->TestRes el
testBoolLatticeBy l els1 els2 els3 targetEls = 
    testComplBy l els1 . testLatticeBy (compLat_lat l) els2 els3 els1 targetEls

testBoolLattice :: (Eq el) => CompLat el -> TestRes el
testBoolLattice      l = 
  let els = compLat_elements l
  in testBoolLatticeBy l els (compLat_atomS l) (compLat_mIrredS l) els
\end{code}
With the standard variant we again automatically reduce the tests to the atoms and %%@
meet-irreducible elements, respectively, to test the lattice structure. 

Finally, our attention goes to morphism tests. According to Definition %%@
\ref{DefnHomomorph}, we differentiate between (upper/lower semi) lattice and co-lattice %%@
morphisms, respectively. Furthermore, morphisms between Brouwerian/Boolean lattices, %%@
which respect the relative pseudo complements/complements are of interest. Obviously, all %%@
kinds of morphisms have in common that they respect unary or binary operations within %%@
their domain and range, respectively. Hence, we can introduce two functions as follows.
\begin{code}
testMorphBy :: (Eq el2) => PoSet el1 -> PoSet el2 -> 
             (el1 -> el1 -> el1) -> 
             (el2 -> el2 -> el2) -> String -> (el1 -> el2) -> TestRes el1
testMorphBy l1 l2 opL1 opL2 text f = 
    foldF [let x1 = head xs
               x2 = last xs
           in testP ((f $ opL1 x1 x2) == (opL2 (f x1) $f x2)) xs text
           | xs <- takeN 2 $ poSet_elements l1]

testMorphUnBy :: (Eq el2) => PoSet el1 -> PoSet el2 -> 
                          (el1 -> el1) -> (el2 -> el2) -> 
                          String -> (el1 -> el2) -> TestRes el1
testMorphUnBy l1 l2 opL1 opL2 text f = 
    foldF [testP ((f $ opL1 x) == (opL2 $ f x)) [x] text | x <- poSet_elements l1]
\end{code}
In both functions \smalltt{f} is the morphism in question. The only difference is that %%@
\smalltt{testMorphBy} gets binary operations (e.g., $sup$, $inf$) and %%@
\smalltt{testMorphUnBy} gets unary (e.g., complement). The string \smalltt{text} is used %%@
to generate an appropriate output when a test fails.

With help of these two functions, we now can easily provide the required tests.
\begin{code}
testUpSemiLatMorph :: (Eq el2) => 
                      USemiLat el1 -> USemiLat el2 -> (el1 -> el2) -> TestRes el1
testUpSemiLatMorph   l1 l2   = 
    testMorphBy (uSemiLat_poSet l1) (uSemiLat_poSet l2) 
              (uSemiLat_sup   l1) (uSemiLat_sup   l2) "USemiLatMorph:"

testLoSemiLatMorph :: (Eq el2) => 
                      LSemiLat el1 -> LSemiLat el2 -> (el1 -> el2) -> TestRes el1
testLoSemiLatMorph   l1 l2   = 
    testMorphBy (lSemiLat_poSet l1) (lSemiLat_poSet l2)
              (lSemiLat_inf   l1) (lSemiLat_inf   l2) "LSemiLatMorph:"

testLatMorph       :: (Eq el2) =>
                      Lat el1 -> Lat el2 -> (el1 -> el2) -> TestRes el1
testLatMorph         l1 l2 f = 
    testUpSemiLatMorph (lat_uSemiLat l1) (lat_uSemiLat l2) f. 
    testLoSemiLatMorph (lat_lSemiLat l1) (lat_lSemiLat l2) f
                            
testUpCoSemiLatMorph :: (Eq el2) => 
                        USemiLat el1 -> LSemiLat el2 -> (el1 -> el2) -> TestRes el1
testUpCoSemiLatMorph l1 l2   = 
    testMorphBy (uSemiLat_poSet l1) (lSemiLat_poSet l2)
              (uSemiLat_sup   l1) (lSemiLat_inf   l2) "UCoSemiLatMorph:"

testLoCoSemiLatMorph :: (Eq el2) => 
                        LSemiLat el1 -> USemiLat el2 -> (el1 -> el2) -> TestRes el1
testLoCoSemiLatMorph l1 l2   = 
    testMorphBy (lSemiLat_poSet l1) (uSemiLat_poSet l2)
              (lSemiLat_inf   l1) (uSemiLat_sup   l2) "LCoSemiLatMorph:" 

testCoLatMorph :: (Eq el2) => Lat el1 -> Lat el2 -> (el1 -> el2) -> TestRes el1
testCoLatMorph       l1 l2 f = 
    testUpCoSemiLatMorph (lat_uSemiLat l1) (lat_lSemiLat l2) f . 
    testLoCoSemiLatMorph (lat_lSemiLat l1) (lat_uSemiLat l2) f

testBoolLatMorph :: (Eq el2) => 
                    CompLat el1 -> CompLat el2 -> (el1 -> el2) -> TestRes el1
testBoolLatMorph     l1 l2 f = 
    testLatMorph  (compLat_lat l1    ) (compLat_lat l2    ) f . 
    testMorphUnBy (compLat_poSet l1  ) (compLat_poSet l2  ) 
                  (compLat_complem l1) (compLat_complem l2) "Complement:" f
\end{code}
Special endomorphisms, namely monotone and antitone functions, can be tested using the %%@
following routines.
\begin{code}
testMonoFunc, testAntiFunc :: PoSet el -> (el -> el) -> TestRes el
testMonoFunc l f = let lEq = poSet_lEq l
                   in foldF [let x1 = head xs
                                 x2 = last xs
                             in testP ((not (lEq x1 x2) || lEq (f x1) (f x2)) &&
                                       (not (lEq x2 x1) || lEq (f x2) (f x1)) )
                             xs "Func.Mono :" | xs <- takeN 2 $ poSet_elements l ]

testAntiFunc l f = let lEq = poSet_lEq l
                   in foldF [let x1 = head xs
                                 x2 = last xs
                             in testP ((not (lEq x1 x2) || lEq (f x2) (f x1)) &&
                                       (not (lEq x2 x1) || lEq (f x1) (f x2)) )
                             xs "Func.Anti :" | xs <- takeN 2 $ poSet_elements l ]                             
\end{code}
Obviously, the tests are done componentwise. Notice that the weakest structure in which %%@
the two notions above make sense are posets.

Thus, our module \smalltt{Lattice} is complete and should suffice for our purposes. The %%@
only thing left to do, is to connect the class view and the module view by building %%@
suitable instances. This is done within the module \smalltt{LatticeInstances}. A listing %%@
can be found in Appendix \ref{AppLatInst}.

