\section{ A module collection for lattices }
\label{SecLatModul}
As we are going to treat $\SL$-fuzzy relations within the algebraic structure of Goguen categories, %%@
we need a module collection for lattices. The representation of an $\SL$-fuzzy relation $R$ by a %%@
mapping $R : A\times B\lora\SL$ makes this even clearer. But, there is another connection. From the %%@
definitions of the relational categories of Section \ref{SecCategoryTheory} we see that the morphism %%@
classes between two objects of a given relational category (except the pure category) form at least a %%@
lower semilattice. Thus, the morphisms of a Dedekind category, for example, constitute a complete %%@
Brouwerian lattice. We do not aim at a comprehensive package for lattices, but provide the most %%@
necessary operations like the determination of the irreducible elements and atoms, respectively. %%@
Another important point will be the consistency tests for lattices and lattice morphisms. Always %%@
consider that we see this module collection as an ``add on'' to RATH. Thus, many design decisions are %%@
made with respect to conformance. Nevertheless, we do not use any functionality of RATH within these %%@
modules so that they can be used as a stand-alone package.

We give two different approaches. The first one uses multi parameter type classes which allows an %%@
elegant parametrization and typing but has the drawback that it is no Haskell 98 standard. The second %%@
one uses record data structures (Haskell 98 standard) as a compensation of Haskell's relatively weak %%@
module system. Thus, we have consistency in using our lattice modules together with RATH. 

First, we want to have a look at the type classes for lattices. In accordance to Definition %%@
\ref{DefnLattice} and Theorem \ref{ThmAlternLattice}, two approaches are possible. We choose to %%@
provide a hierarchy starting with partially ordered sets and then add lattice operations %%@
successively. Of course, the induced ordering and join/meet can be mutually computed by standard %%@
operations, but we let the user the chance to define more efficient operations.
\begin{code}
module LatticeClass where

class POrderedSet set el where
    isElem    :: set -> el -> Bool      
    elements  :: set -> [el]   
    lEq       :: set -> el -> el -> Bool
\end{code}
As we are going to use our lattice module together with RATH later on, we have to take 
care of the name spaces to avoid longish qualified names. Thus, we speak of elements 
rather than of objects. Furthermore, the ordering is denoted by ``lower or equal'' (\smalltt{lEq}) 
instead of \smalltt{incl}. 

The next step is to provide the sup and inf operation, respectively. This results in upper 
resp.\ lower semilattices. 
\begin{code}
class POrderedSet lat el => UpSemiLattice lat el where
    sup     :: lat -> el -> el -> el 
    topEl   :: lat -> el 

class POrderedSet lat el => LoSemiLattice lat el where
    inf      :: lat -> el -> el -> el  
    botEl    :: lat -> el                
    atomS    :: lat -> [el]
\end{code}  
We aim to compute finite (thus, complete) lattices. According to Lemma \ref{Lem01Lattice} we %%@
therefore have a least and a greatest element, respectivly, denoted by \smalltt{botEl} and %%@
\smalltt{topEl}, respectively. Furthermore, every non-trivial lower semilattice with least element %%@
contains at least one atom. As the atoms reflect the whole structure in some kinds of lattices (e.g. %%@
Boolean, cf.\ Section \ref{SubSecBoolLat}), they play an important role with the consistency tests %%@
later on. 

The next step is a type class for (standard) lattices. 
\begin{code}
class (UpSemiLattice lat el, LoSemiLattice lat el) => Lattice lat el where 
    jIrredS  :: lat -> [el]
    mIrredS  :: lat -> [el]
\end{code}
We do not have to provide any new operations. But, in a finite lattice, all elements can be computed %%@
by the disjunction/conjunction of a finite number of join-/meet-irreducible elements (cf.\ Theorem %%@
\ref{ThmDisjIrred}, Definition \ref{DefnIrred}). Thus, every non-trivial lattice with least/greatest %%@
element contains at least one join-/meet-irreducible element. Again, irreducible elements will play %%@
an important role with the consistency tests and several applications later on. 

As we saw in the section about relational categories, Brouwerian lattices are essential. Within this %%@
kind of lattice, we have a new operation for the computation of the relative pseudo complements. %%@
Therefore, we define the following type class. 
\begin{code}
class (LoSemiLattice lat el) => RelCompLattice lat el where
    relComplem :: lat -> el -> el -> el
\end{code}
Notice, that we deliberately do not call the type class ``BrouwerianLattice'' or similar because the %%@
operation \smalltt{relComplem} can be interesting even if the relative pseudo complements do not %%@
exist for {\em every} pair of elements. We even do not demand that \smalltt{lat} and \smalltt{el} %%@
form a lattice because this is not necessary for the definition. 

The even stronger notion of a complement (cf.\ Definition \ref{DefnCompl}) is taken into account by %%@
the next type class. 
\begin{code}
class Lattice compLat el => CompLattice compLat el where
    complem :: compLat -> el -> el
\end{code}  
Again, we avoid designations like ``BooleanLattice''.

Hence, we have the type class hierarchy shown in Figure \ref{LatticeTypeClasses}.\\
\begin{figure}[h]
\begin{center}
\begin{picture}(300,230)
\put(100,5  ){\framebox(100,15){\smalltt{POrderedSet}}}
\put(5,75   ){\framebox(100,15){\smalltt{LoSemiLattice}}}
\put(195,75 ){\framebox(100,15){\smalltt{UpSemiLattice}}}
\put(110,145){\framebox(80,15 ){\smalltt{Lattice}}}
\put(10,145 ){\framebox(90,15 ){\smalltt{RelCompLattice}}}
\put(100,215){\framebox(100,15){\smalltt{CompLattice}}}
\put(110,20 ){\line(-1,1){55}}
\put(190,20 ){\line(1,1 ){55}}
\put(55,90  ){\line(1,1 ){55}}
\put(245,90 ){\line(-1,1){55}}
\put(55,90  ){\line(0,1 ){55}}
\put(150,160){\line(0,1 ){55}}
\end{picture}
\end{center}
\caption{\label{LatticeTypeClasses} The hierarchy of the lattice type classes}
\end{figure}\\
