\section { Using the lattice module }
Now, we want to demonstrate how the module of the last section can be used to instantiate %%@
some standard lattices. With these remarks it then should be clear how the module works. %%@
Throughout this section, we only instantiate data structures that form lattices or %%@
complementary lattices. The corresponding (semi)lattice structures then can be extracted %%@
using the standard routines of the last section. 
\begin{code}
module StandardLattices where
import Lattice
import List (delete,nub,union)
import Sets
\end{code}
First, we want to create the complete Boolean lattice induced by the powerset of a given %%@
set. To do so, we import the \smalltt{Sets} module. Unfortunately, this module provides %%@
no \smalltt{Eq} instance for the \smalltt{Set} data structure. Since we need this, we %%@
have to implement our own instance.
\begin{code}
instance (Eq obj,Ord obj) => Eq (Set obj) where
  (==) s1 s2 = let ls1 = sizeSet s1
               in ls1 == (sizeSet s2) &&
                  ls1 == (sizeSet $ intersectSet s1 s2)
\end{code}
Furthermore, we need a function that, given a set, returns the powerset.
\begin{code}
power :: [a] -> [[a]]
power l = power' id l []
 where
  power' f [] = ((f []):)
  power' f (x:xs) = power' f xs . power' (f . (x:)) xs
\end{code}
With this sets are represented by a simple element list. Hence, we are able to %%@
instantiate the induced Boolean lattice using the \smalltt{Set} data structure.
\begin{code}
powerSetLat :: (Eq el,Ord el) => [el] -> CompLat (Set el)
powerSetLat els = 
          let els' = map listToSet $ power els
              ats  = map (listToSet.(:[])) $ nub els
              tEl  = listToSet els
          in  CompLat { compLat_lat    = 
                          Lat { lat_poSet   = 
                                 PoSet { poSet_isElem   = flip elem $ els'
                                        ,poSet_elements = els'
                                        ,poSet_lEq      = \x y -> intersectSet 
                                                                  x y == x }
                               ,lat_sup     = joinSet
                               ,lat_inf     = intersectSet
                               ,lat_topEl   = tEl
                               ,lat_botEl   = zeroSet
                               ,lat_atomS   = ats
                               ,lat_jIrredS = ats
                               ,lat_mIrredS = map (diffSet tEl) ats }
                        ,compLat_complem = diffSet tEl }
\end{code}
The implementation is straightforward and mainly uses the predefined functions of the %%@
\smalltt{Sets} module. Notice that we explicitly provide the atoms and irreducible %%@
elements, respectively, to avoid unnecessary computation times.\\
Now, we have to check whether the lattice is instantiated correctly.
First, we exemplary want to have a look at the created elements and the atoms of the %%@
powerset lattice induced by the set $\{1,2,3,4\}$. The Hugs session
\begin{smallverb}
StandardLattices> compLat_elements $ powerSetLat [1..4]
[{},{4},{3},{3, 4},{2},{2, 4},{2, 3},{2, 3, 4},{1},
 {1, 4},{1, 3},{1, 3, 4},{1, 2},{1, 2, 4},{1, 2, 3},{1, 2, 3, 4}]

StandardLattices> compLat_atomS $ powerSetLat [1..4]
[{1},{2},{3},{4}]
\end{smallverb}
shows that these elements are generated correctly. The same is true for all other parts %%@
of the data structure. Calling the consistency tests
\begin{smallverb}
StandardLattices> testBoolLattice (powerSetLat [1..4]) []
[]

StandardLattices> testAtomic (compLat_lat $ powerSetLat [1..4]) [] 
[]
\end{smallverb}
makes sure that we indeed have a correctly instantiated Boolean lattice.

The second standard Boolean lattice consists of the truth values with implication $\ra$ %%@
as the ordering and $True$ resp.\ $False$ as greatest resp.\ least element. It is the %%@
base for constructing Boolean matrices using the \smalltt{LFuzzyRel} module.
\begin{code}
logicLat :: CompLat Bool
logicLat = let els = [True,False]
           in  CompLat { compLat_lat    = 
                          Lat { lat_poSet   = 
                                 PoSet { poSet_isElem   = flip elem $ els
                                        ,poSet_elements = els
                                        ,poSet_lEq      = \x y -> not x || y }
                               ,lat_sup     = (||)
                               ,lat_inf     = (&&)
                               ,lat_topEl   = True
                               ,lat_botEl   = False
                               ,lat_atomS   = [True]
                               ,lat_jIrredS = [True]
                               ,lat_mIrredS = [False] }
                        ,compLat_complem = not }

\end{code}
Obviously, this definition is consistent.

Another important class are linear lattices (i.e., $a\leq b$ or $b\leq a$ holds for every %%@
two elements). The unit interval $[0,1]$, for instance, constitutes such a lattice. We %%@
provide the function 
\begin{code}
linLat :: Ord el => [el] -> Lat el
linLat els = let
               mi    = minimum els
               ma    = maximum els
               notMi = delete mi els
             in Lat { lat_poSet   = PoSet { poSet_isElem   = flip elem $ els
                                           ,poSet_elements = els
                                           ,poSet_lEq      = (<=) }
                     ,lat_sup     = max
                     ,lat_inf     = min
                     ,lat_topEl   = ma
                     ,lat_botEl   = mi
                     ,lat_atomS   = [minimum notMi]
                     ,lat_jIrredS = notMi
                     ,lat_mIrredS = delete ma els }
\end{code}
which instantiates the lattice in question. Notice that all elements except the least are %%@
join-irreducible and all elements except the greatest are meet-irreducible. Again, %%@
typing, for example
\begin{smallverb}
StandardLattices> testLattice (linLat [1,2,3,4,6,12]) []
[] 
\end{smallverb}
assures the consistency.

The next construction is of a somehow mathematical nature. We provide the lattice induced %%@
by the divisibility relation with \smalltt{lcm} and \smalltt{gcd} as supremum resp.\ %%@
infimum operation.
\begin{code}
divLat :: Int -> Int -> Lat Int
divLat lB uB = let
                 els = [ x | x<-[lB..uB],mod uB x == 0]
                 mi  = minimum els
                 pS  = PoSet { poSet_isElem   = flip elem $ els
                              ,poSet_elements = els
                              ,poSet_lEq      = \x y -> mod y x == 0 }
               in  Lat { lat_poSet   = pS
                        ,lat_sup     = lcm
                        ,lat_inf     = gcd
                        ,lat_topEl   = uB
                        ,lat_botEl   = mi
                        ,lat_atomS   = atomSetBy pS mi gcd []
                        ,lat_jIrredS = irredSetBy pS mi lcm []
                        ,lat_mIrredS = irredSetBy pS uB gcd [] }
\end{code}
Since we are dealing with finite structures, the user has to give a lower and upper %%@
bound, respectively, which delimit the possible elements of the lattice. We take %%@
\smalltt{uB} as the greatest element so that \smalltt{lB} is not necessarily the least %%@
element of the resulting lattice. It is the least element, if and only if it is a divisor %%@
of \smalltt{uB}. Notice that the atoms of the resulting lattice are exactly the prime %%@
factors of \smalltt{uB} if \smalltt{lB} is $1$.

Finally, we want to implement the non-modular and non-distributive lattice. They are %%@
interesting in the way that we can compare whether the modularity resp.\ distributivity %%@
check fail.
\begin{code}
nonModLat, nonDistrLat :: Lat String
nonModLat = 
  let els = ["0","a","b","c","1"]
      irr = ["a","b","c"]
  in Lat { lat_poSet   = PoSet { poSet_isElem   = flip elem $ els
                                ,poSet_elements = els
                                ,poSet_lEq      = \x y -> x=="0" || y=="1" ||
                                                     x==y || (x=="a" && y=="b") }
          ,lat_sup     = \x y -> if x=="0" then y else 
                                   if y=="0" then x else
                                     if x==y then x else 
                                       if (x=="a")&&(y=="b")||
                                          (x=="b")&&(y=="a") then "b" else "1"
          ,lat_inf     = \x y -> if x=="1" then y else
                                   if y=="1" then x else
                                     if x==y then x else 
                                       if (x=="a")&&(y=="b")||
                                          (x=="b")&&(y=="a") then "a" else "0"
           ,lat_topEl   = "1"
           ,lat_botEl   = "0"
           ,lat_atomS   = ["a","c"]
           ,lat_jIrredS = irr
           ,lat_mIrredS = irr }

nonDistrLat = 
  let els = ["0","a","b","c","1"]
      ats = ["a","b","c"]
  in Lat { lat_poSet   = PoSet { poSet_isElem   = flip elem $ els
                                ,poSet_elements = els
                                ,poSet_lEq      = \x y -> x=="0" || y=="1" || 
                                                          x==y }
           ,lat_sup     = \x y -> if x=="0" then y else 
                                    if y=="0" then x else
                                      if x==y then x else "1"
           ,lat_inf     = \x y -> if x=="1" then y else
                                    if y=="1" then x else
                                      if x==y then x else "0"
           ,lat_topEl   = "1"
           ,lat_botEl   = "0"
           ,lat_atomS   = ats
           ,lat_jIrredS = ats
           ,lat_mIrredS = ats }
\end{code}
The implementations directly correspond to the lattices shown in Figure %%@
\ref{NoModularLatt} and \ref{NoDistrLatt}. 
The Hugs session
\begin{smallverb}
StandardLattices> testModular nonModLat []
[("Modularity:",["a","c","b"])]

StandardLattices> testModular nonDistrLat []
[]

StandardLattices> testJDistr nonDistrLat []
[("Distributivity:",["a","b","c"])]

StandardLattices> testMDistr nonDistrLat []
[("Distributivity:",["a","b","c"])]  
\end{smallverb}
indeed shows that \smalltt{nonModLat} is not modular and \smalltt{nonDistrLat} is modular %%@
but not distributive.%%@
